from datetime import datetime

from .constants import EResult


class SteamError(Exception):
    """All errors related to Steam"""


class EResultError(SteamError):
    """Raised when Steam response data contain `success` field with error code"""

    def __init__(self, msg: str, result: EResult, data=None):
        self.msg = msg
        self.result = result
        self.data = data

    def __str__(self):
        return self.msg


class LoginError(SteamError):
    """Raised when a problem with login process occurred"""


# TODO What about
#  https://github.com/DoctorMcKay/node-steamcommunity/blob/1067d4572ee9d467e8f686951901c51028c5c995/components/http.js#L94
class SessionExpired(SteamError):
    """Raised when session is expired, and you need to do login"""


class RateLimitExceeded(SteamError):
    """Raised when Steam decided you were in need of a bit of a rest :)"""


class ResourceNotModified(SteamError):
    """
    Special case when `If-Modified-Since` header included
    in request headers and Steam response with 304 status code
    """

    def __init__(self, last_modified: datetime, expires: datetime):
        self.last_modified = last_modified
        self.expires = expires

    def __str__(self):
        return ""


# class InsufficientBalance(SteamError):
#     """"""



"""Typed dicts for responses and methods."""

from typing import TypedDict, Literal, TypeAlias


# I am too lazy to make enum
PrivacySettingsOptions: TypeAlias = Literal[1, 2, 3]
PrivacySettingsCheckboxOptions: TypeAlias = Literal[1, 3]
CommentPrivacySettingsOptions: TypeAlias = Literal[0, 1, 2]


class SellOrderTableData(TypedDict):
    price: str
    price_with_fee: str
    quantity: str


class BuyOrderTableData(TypedDict):
    price: str
    quantity: str


class ItemOrdersHistogramData(TypedDict):
    success: int
    sell_order_count: str
    sell_order_price: str
    sell_order_table: list[SellOrderTableData]
    buy_order_count: str
    buy_order_price: str
    buy_order_table: list[BuyOrderTableData]
    highest_buy_order: str
    lowest_sell_order: str

    # actually there is a lists, but tuple typing have fixed values
    buy_order_graph: list[tuple[float, int, str]]
    sell_order_graph: list[tuple[float, int, str]]

    graph_max_y: int
    graph_min_x: float
    graph_max_x: float
    price_prefix: str
    price_suffix: str


class Activity(TypedDict):
    type: str
    quantity: str
    price: str
    time: int
    avatar_buyer: str
    avatar_medium_buyer: str
    persona_buyer: str
    avatar_seller: str
    avatar_medium_seller: str
    persona_seller: str


class ItemOrdersActivity(TypedDict):
    success: int
    activity: list[Activity]
    timestamp: int


class PriceOverview(TypedDict):
    success: int
    lowest_price: str
    volume: str
    median_price: str


class TradeOffersSummary(TypedDict):
    pending_received_count: int
    new_received_count: int
    updated_received_count: int
    historical_received_count: int
    pending_sent_count: int
    newly_accepted_sent_count: int
    updated_sent_count: int
    historical_sent_count: int
    escrow_received_count: int
    escrow_sent_count: int


class WalletInfo(TypedDict):
    wallet_currency: int
    wallet_country: str
    wallet_state: str
    wallet_fee: str
    wallet_fee_minimum: str
    wallet_fee_percent: str
    wallet_publisher_fee_percent_default: str
    wallet_fee_base: str
    wallet_balance: str
    wallet_delayed_balance: str
    wallet_max_balance: str
    wallet_trade_max_balance: str
    success: int
    rwgrsn: int


class UserWallet(TypedDict):
    amount: str  # int
    currency: str


class FundWalletInfo(TypedDict):
    """From `https://store.steampowered.com/api/getfundwalletinfo`"""

    success: int
    currency: str
    country_code: str
    alternate_min_amount: bool
    amounts: list[int]
    related_trans_type: bool
    related_trainsid: bool
    user_wallet: UserWallet


class JWTToken(TypedDict):
    iss: str
    sub: str  # steam id64
    aud: list[str]
    exp: int
    nbf: int
    iat: int
    jti: str
    oat: int
    per: int
    ip_subject: str
    ip_confirmer: str


class LocationData(TypedDict):
    locCity: str
    locCityCode: int
    locCountry: str
    locCountryCode: str
    locState: str
    locStateCode: str


class ProfilePrivacySettings(TypedDict):
    PrivacyFriendsList: PrivacySettingsOptions
    PrivacyInventory: PrivacySettingsOptions
    PrivacyInventoryGifts: PrivacySettingsCheckboxOptions
    PrivacyOwnedGames: PrivacySettingsOptions
    PrivacyPlaytime: PrivacySettingsCheckboxOptions
    PrivacyProfile: PrivacySettingsOptions


class ProfilePrivacy(TypedDict):
    PrivacySettings: ProfilePrivacySettings
    eCommentPermission: CommentPrivacySettingsOptions


class ProfilePreferences(TypedDict):
    hide_profile_awards: int


class ProfileData(TypedDict):
    strPersonaName: str
    strCustomURL: str
    strRealName: str
    strSummary: str
    strAvatarHash: str
    rtPersonaNameBannedUntil: str
    rtProfileSummaryBannedUntil: str
    rtAvatarBannedUntil: str
    LocationData: LocationData
    # ActiveTheme
    ProfilePreferences: ProfilePreferences
    # rgAvailableThemes
    # rgGoldenProfileData
    Privacy: ProfilePrivacy


class AvatarUploadImagesData(TypedDict):
    # 0: str  illegal variable annotation, funny
    medium: str
    full: str


class AvatarUploadData(TypedDict):
    hash: str
    images: AvatarUploadImagesData
    message: str



"""Abstract utils within `Steam` context and not"""

import asyncio
from base64 import b64decode, b64encode
from datetime import datetime
from struct import pack, unpack
from time import time as time_time
from hmac import new as hmac_new
from hashlib import sha1
from functools import wraps, partial
from typing import Callable, overload, TypeVar, TypeAlias, Literal
from http.cookies import SimpleCookie, Morsel
from math import floor
from secrets import token_hex
from re import search as re_search, compile as re_compile
from json import loads as j_loads

from aiohttp import ClientSession, ClientResponse
from yarl import URL

from .typed import JWTToken


__all__ = (
    "gen_two_factor_code",
    "generate_confirmation_key",
    "generate_device_id",
    "extract_openid_payload",
    "do_session_steam_auth",
    "get_cookie_value_from_session",
    "remove_cookie_from_session",
    "async_throttle",
    "create_ident_code",
    "account_id_to_steam_id",
    "steam_id_to_account_id",
    "id64_to_id32",
    "id32_to_id64",
    "to_int_boolean",
    "get_jsonable_cookies",
    "update_session_cookies",
    "buyer_pays_to_receive",
    "receive_to_buyer_pays",
    "generate_session_id",
    "decode_jwt",
    "find_item_nameid_in_text",
    "patch_session_with_http_proxy",
    "parse_time",
    "format_time",
    "attribute_required",
    "make_inspect_url",
    "add_cookie_to_session",
    "calc_market_listing_fee",
)


def gen_two_factor_code(shared_secret: str, timestamp: int = None) -> str:
    """Generate twofactor (onetime/TOTP) code."""

    if timestamp is None:
        timestamp = int(time_time())
    time_buffer = pack(">Q", timestamp // 30)  # pack as Big endian, uint64
    time_hmac = hmac_new(b64decode(shared_secret), time_buffer, digestmod=sha1).digest()
    begin = ord(time_hmac[19:20]) & 0xF
    full_code = unpack(">I", time_hmac[begin : begin + 4])[0] & 0x7FFFFFFF  # unpack as Big endian uint32
    chars = "23456789BCDFGHJKMNPQRTVWXY"
    code = ""

    for _ in range(5):
        full_code, i = divmod(full_code, len(chars))
        code += chars[i]

    return code


def generate_confirmation_key(identity_secret: str, tag: str, timestamp: int = None) -> str:
    if timestamp is None:
        timestamp = int(time_time())
    buff = pack(">Q", timestamp) + tag.encode("ascii")
    return b64encode(hmac_new(b64decode(identity_secret), buff, digestmod=sha1).digest()).decode()


# It works, however it's different that one generated from mobile app
def generate_device_id(steam_id: int) -> str:
    """Generate mobile android device id. Confirmation endpoints requires this."""

    hexed_steam_id = sha1(str(steam_id).encode("ascii")).hexdigest()
    return "android:" + "-".join(
        [hexed_steam_id[:8], hexed_steam_id[8:12], hexed_steam_id[12:16], hexed_steam_id[16:20], hexed_steam_id[20:32]]
    )


def extract_openid_payload(page_text: str) -> dict[str, str]:
    """
    Extract steam openid payload (specs) from page html raw text.
    Use it if 3rd party websites have extra or non-cookie auth (JWT via service API call, for ex.).

    :param page_text:
    :return: dict with payload data
    """

    # not so beautiful as with bs4 but dependency free
    return {
        "action": re_search(r"id=\"actionInput\"[\w=\"\s]+value=\"(?P<action>\w+)\"", page_text)["action"],
        "openid.mode": re_search(r"name=\"openid\.mode\"[\w=\"\s]+value=\"(?P<mode>\w+)\"", page_text)["mode"],
        "openidparams": re_search(r"name=\"openidparams\"[\w=\"\s]+value=\"(?P<params>[\w=/]+)\"", page_text)["params"],
        "nonce": re_search(r"name=\"nonce\"[\w=\"\s]+value=\"(?P<nonce>\w+)\"", page_text)["nonce"],
    }


async def do_session_steam_auth(session: ClientSession, auth_url: str | URL) -> ClientResponse:
    """
    Request auth page, find specs of steam openid and log in through steam with passed session.
    Use it when you need to log in 3rd party site trough Steam using only cookies.

    .. seealso:: https://aiosteampy.somespecial.one/examples/auth_3rd_party_site/

    :param session: just session.
    :param auth_url: url to site, which redirect you to steam login page.
    :return: response with history, headers and data
    """

    r = await session.get(auth_url)
    rt = await r.text()

    data = extract_openid_payload(rt)

    return await session.post("https://steamcommunity.com/openid/login", data=data, allow_redirects=True)


def get_cookie_value_from_session(session: ClientSession, url: URL | str, field: str) -> str | None:
    """Get value from session cookies. Passed `url` must include scheme (for ex. `https://url.com`)."""

    c = session.cookie_jar.filter_cookies(URL(url))
    return c[field].value if field in c else None


def remove_cookie_from_session(session: ClientSession, url: URL | str, field: str) -> bool:
    """Remove cookie from session cookies. Return `True` if cookie was present and removed."""

    raw = str(url)
    if "//" in raw:
        host = raw.split("//")[1]
    else:
        host = raw
    return bool(session.cookie_jar._cookies[(host, "/")].pop(field, None))


_R = TypeVar("_R")


@overload
def async_throttle(seconds: float, *, arg_index: int) -> Callable[[Callable[..., _R]], Callable[..., _R]]:
    ...


@overload
def async_throttle(seconds: float, *, arg_name: str) -> Callable[[Callable[..., _R]], Callable[..., _R]]:
    ...


@overload
def async_throttle(seconds: float) -> Callable[[Callable[..., _R]], Callable[..., _R]]:
    ...


def async_throttle(seconds, *, arg_index=None, arg_name=None):
    """
    Prevents the decorated function from being called more than once per `seconds`.
    Throttle (`await asyncio.sleep`) before call wrapped async func,
    when related arg was equal (same hash value) to related arg passed in previous func call
    if time between previous and current call lower than `seconds`.
    Related arg must be hashable (can be dict key).
    Wrapped func must be async (return `Coroutine`).
    Throttle every func call time if related arg has been not specified.

    :param seconds: seconds during which call frequency has been limited.
    :param arg_index: index of related arg in *args tuple.
    :param arg_name: keyname of related arg in **kwargs.
    """

    # mega optimization, prevent arg checks in wrapped func call
    # I know about PEP8: E731, but this way is much shorter and readable
    if arg_index is None and arg_name is None:
        get_key = lambda _, __: None
    elif arg_index:
        get_key = lambda a, _: a[arg_index]
    else:
        get_key = lambda _, k: k[arg_name]

    ts_map: dict[..., float] = {}

    def decorator(f: Callable[..., _R]) -> Callable[..., _R]:
        @wraps(f)
        async def wrapper(*args, **kwargs) -> _R:
            key = get_key(args, kwargs)
            ts = time_time()
            diff = ts - ts_map.get(key, 0)
            if diff < seconds:
                await asyncio.sleep(diff + 0.01)  # +10ms to ensure that second passed since last call
                ts_map[key] = time_time()
            else:
                ts_map[key] = ts

            return await f(*args, **kwargs)

        return wrapper

    return decorator


@overload
def create_ident_code(asset_id: int | str, context_id: int | str, app_id: int | str, *, sep: str = ...) -> str:
    ...


@overload
def create_ident_code(instance_id: int | str, class_id: int | str, app_id: int | str, *, sep: str = ...) -> str:
    ...


def create_ident_code(*args, sep=":"):
    """
    Create unique ident code for `EconItem` asset or `ItemDescription` within whole `Steam Economy`.

    .. seealso:: https://dev.doctormckay.com/topic/332-identifying-steam-items/
    """

    return sep.join(reversed(list(str(i) for i in filter(lambda i: i is not None, args))))


def steam_id_to_account_id(steam_id: int) -> int:
    """Convert steam id64 to steam id32."""

    return steam_id & 0xFFFFFFFF


def account_id_to_steam_id(account_id: int) -> int:
    """Convert steam id32 to steam id64."""

    return 1 << 56 | 1 << 52 | 1 << 32 | account_id


id64_to_id32 = steam_id_to_account_id
id32_to_id64 = account_id_to_steam_id


def to_int_boolean(s):
    """Convert something to 1, 0."""

    return 1 if s else 0


JSONABLE_COOKIE_JAR: TypeAlias = list[dict[str, dict[str, str, None, bool]]]


def update_session_cookies(session: ClientSession, cookies: JSONABLE_COOKIE_JAR):
    """Update the session cookies from jsonable cookie jar."""

    for cookie_data in cookies:
        c = SimpleCookie()
        for k, v in cookie_data.items():
            copied = dict(**v)  # copy to avoid modification of the arg
            m = Morsel()
            m._value = copied.pop("value")
            m._key = copied.pop("key")
            m._coded_value = copied.pop("coded_value")
            m.update(copied)
            c[k] = m

        session.cookie_jar.update_cookies(c)


def get_jsonable_cookies(session: ClientSession) -> JSONABLE_COOKIE_JAR:
    """Extract and convert cookies to dict object."""

    return [
        {
            field_key: {
                "coded_value": morsel.coded_value,
                "key": morsel.key,
                "value": morsel.value,
                "expires": morsel["expires"],
                "path": morsel["path"],
                "comment": morsel["comment"],
                "domain": morsel["domain"],
                "max-age": morsel["max-age"],
                "secure": morsel["secure"],
                "httponly": morsel["httponly"],
                "version": morsel["version"],
                "samesite": morsel["samesite"],
            }
            for field_key, morsel in cookie.items()
        }
        for cookie in session.cookie_jar._cookies.values()
    ]


def receive_to_buyer_pays(
    amount: int,
    *,
    publisher_fee=0.1,
    steam_fee=0.05,
    wallet_fee_min=1,
    wallet_fee_base=0,
) -> tuple[int, int, int]:
    """
    Convert `to_receive` amount in `buyer_pays`. Mostly needed for placing sell listing.
    Works just like on sell listing window on steam.

    :param amount: amount in cents
    :return: steam fee value, publisher fee value, buyer pays amount
    """

    steam_fee_value = int(floor(max(amount * steam_fee, wallet_fee_min) + wallet_fee_base))
    publisher_fee_value = int(floor(max(amount * publisher_fee, 1) if publisher_fee > 0 else 0))
    return steam_fee_value, publisher_fee_value, int(amount + steam_fee_value + publisher_fee_value)


def buyer_pays_to_receive(
    amount: int,
    *,
    publisher_fee=0.1,
    steam_fee=0.05,
    wallet_fee_min=1,
    wallet_fee_base=0,
) -> tuple[int, int, int]:
    """
    Convert `buyer_pays` amount in `to_receive`. Mostly needed for placing sell listing.
    Works just like on sell listing window on steam.

    :param amount: amount in cents
    :return: steam fee value, publisher fee value, amount to receive
    """

    # I don't know how it works, it's just a copy of js function working on inputs in steam front
    estimated_amount = int((amount - wallet_fee_base) / (steam_fee + publisher_fee + 1))
    s_fee, p_fee, v = receive_to_buyer_pays(
        estimated_amount,
        publisher_fee=publisher_fee,
        steam_fee=steam_fee,
        wallet_fee_min=wallet_fee_min,
        wallet_fee_base=wallet_fee_base,
    )

    i = 0
    some_flag = False
    while (v != amount) and (i < 10):
        if v > amount:
            if some_flag:
                s_fee, p_fee, v = receive_to_buyer_pays(
                    estimated_amount - 1,
                    publisher_fee=publisher_fee,
                    steam_fee=steam_fee,
                    wallet_fee_min=wallet_fee_min,
                    wallet_fee_base=wallet_fee_base,
                )
                s_fee += amount - v
                v = amount
                break
            else:
                estimated_amount -= 1
        else:
            some_flag = True
            estimated_amount += 1

        s_fee, p_fee, v = receive_to_buyer_pays(
            estimated_amount,
            publisher_fee=publisher_fee,
            steam_fee=steam_fee,
            wallet_fee_min=wallet_fee_min,
            wallet_fee_base=wallet_fee_base,
        )
        i += 1

    return s_fee, p_fee, int(v - s_fee - p_fee)


# https://github.com/DoctorMcKay/node-steam-session/blob/698469cdbad3e555dda10c81f580f1ee3960156f/src/LoginSession.ts#L801C19-L801C50
def generate_session_id() -> str:
    """Generate steam like session id."""

    # Hope ChatGPT knows what she is doing
    return token_hex(12)


def decode_jwt(token: str) -> JWTToken:
    parts = token.split(".")
    if len(parts) != 3:
        raise ValueError("Invalid JWT", parts)

    return j_loads(b64decode(parts[1] + "==", altchars="-_"))


_ITEM_NAMEID_RE = re_compile(r"Market_LoadOrderSpread\(\s?(?P<nameid>\d+)\s?\)")


def find_item_nameid_in_text(text: str) -> int | None:
    """Find and return`item_nameid` in HTML text response from `Steam Community Market` page"""

    res = _ITEM_NAMEID_RE.search(text)
    return int(res["nameid"]) if res is not None else res


def patch_session_with_http_proxy(session: ClientSession, proxy: str | URL) -> ClientSession:
    """Patch `aiohttp.ClientSession` to make all requests go through web proxy"""

    session._request = partial(session._request, proxy=proxy)
    return session


_HEADER_TIME_FORMAT = "%a, %d %b %Y %H:%M:%S %Z"


def parse_time(value: str) -> datetime:
    """
    Parse header time (`Last-Modified`, `Expires`, ...),
    cookie `expires` fields to a timezone naive datetime object
    """

    return datetime.strptime(value, _HEADER_TIME_FORMAT)


def format_time(d: datetime) -> str:
    """Format timezone naive datetime object to header/cookie acceptable string"""

    return d.strftime(_HEADER_TIME_FORMAT) + "GMT"  # simple case


# generic, but less performant due to getattr
# without typing, PyCharm complains about return type while VsCode not
def attribute_required(attr: str, msg: str = None):
    """Generate a decorator that check required `attr` on instance before call a wrapped method"""

    def decorator(func):
        @wraps(func)
        def wrapper(self, *args, **kwargs):
            if getattr(self, attr, None) is None:
                raise AttributeError(msg or f"You must provide a value for '{attr}' before using this method")
            return func(self, *args, **kwargs)

        return wrapper

    return decorator


@overload
def make_inspect_url(*, owner_id: int, asset_id: int, d_id: int) -> str:
    ...


@overload
def make_inspect_url(*, market_id: int, asset_id: int, d_id: int) -> str:
    ...


def make_inspect_url(*, market_id: int = None, owner_id: int = None, asset_id: int, d_id: int) -> str:
    if market_id:
        return f"steam://rungame/730/76561202255233023/+csgo_econ_action_preview%20M{market_id}A{asset_id}D{d_id}"
    else:
        return f"steam://rungame/730/76561202255233023/+csgo_econ_action_preview%20S{owner_id}A{asset_id}D{d_id}"


def add_cookie_to_session(
    session: ClientSession,
    url: URL | str,
    name: str,
    value: str,
    *,
    path="/",
    expires: datetime | str = None,
    samesite: str | Literal[True] = None,
    secure: bool = False,
    httponly: bool = False,
):
    if isinstance(url, str):
        url = URL(url)

    c = SimpleCookie()
    c[name] = value
    c[name]["path"] = path
    c[name]["domain"] = url.host
    if expires is not None:
        if isinstance(expires, datetime):
            c[name]["expires"] = format_time(expires)
        else:  # str
            c[name]["expires"] = expires
    if samesite is not None:
        c[name]["samesite"] = samesite
    if secure:
        c[name]["secure"] = secure
    if httponly:
        c[name]["httponly"] = httponly

    session.cookie_jar.update_cookies(cookies=c, response_url=url)


def calc_market_listing_fee(price: int, *, wallet_fee=0.05, publisher_fee=0.10, minimal_fee=1) -> int:
    """
    Calculate market fee for listing.

    Use `get_wallet_info` method of the client to see fees values for specified `Steam` account

    :param price: price of market listing
    :param wallet_fee:
    :param publisher_fee:
    :param minimal_fee: minimal fee value
    :return: calculated fee of price as integer
    """
    return (floor(price * wallet_fee) or minimal_fee) + (floor(price * publisher_fee) or minimal_fee)



from typing import overload, Literal
from json import loads
from re import compile
from datetime import datetime

from ..constants import STEAM_URL, ConfirmationType, AppContext, CORO, EResult
from ..exceptions import EResultError, SessionExpired
from ..models import Confirmation, MyMarketListing, EconItem, TradeOffer
from ..utils import create_ident_code
from .login import LoginMixin


CONF_URL = STEAM_URL.COMMUNITY / "mobileconf"
ITEM_INFO_RE = compile(r"'confiteminfo', (?P<item_info>.+), UserYou")
CONF_OP_TAGS = Literal["allow", "cancel"]


class ConfirmationMixin(LoginMixin):
    """
    Mixin with confirmations related methods. Requires `_identity_secret` to be set.
    Depends on `LoginMixin`.
    """

    __slots__ = ()

    @overload
    async def confirm_sell_listing(self, obj: int, app_context: AppContext) -> Confirmation:
        ...

    @overload
    async def confirm_sell_listing(self, obj: MyMarketListing | EconItem | int) -> Confirmation:
        ...

    async def confirm_sell_listing(
        self,
        obj: MyMarketListing | EconItem | int,
        app_context: AppContext = None,
    ) -> Confirmation:
        """
        Perform sell listing confirmation.
        Pass `game` arg only with asset id.

        :param obj: `MyMarketListing` or `EconItem` that you listed or listing id or asset id
        :param app_context: `Steam` app+context. Required when `obj` is asset id
        :return: `Confirmation`
        """

        update_listings = False  # avoid unnecessary requests
        if isinstance(obj, MyMarketListing):
            key = obj.id
        elif isinstance(obj, EconItem):
            key = obj.id
            update_listings = True
        else:  # int
            if app_context is not None:  # asset id & app
                key = create_ident_code(obj, app_context.context, app_context.app.value)
                update_listings = True
            else:  # listing id
                key = obj

        conf = await self.get_confirmation(key, update_listings=update_listings)
        await self.allow_confirmation(conf)

        return conf

    async def confirm_api_key_request(self, request_id: str) -> Confirmation:
        """Perform api key request confirmation."""

        conf = await self.get_confirmation(request_id)
        await self.allow_confirmation(conf)

        return conf

    async def confirm_trade_offer(self, obj: int | TradeOffer) -> Confirmation:
        """Perform sell trade offer confirmation."""

        conf = await self.get_confirmation(obj.id if isinstance(obj, TradeOffer) else obj)
        await self.allow_confirmation(conf)

        return conf

    async def get_confirmation(self, key: str | int, *, update_listings=True) -> Confirmation:
        """
        Fetch all confirmations from `Steam`, filter and get one.

        :param key: `MarketListingItem` ident code, `TradeOffer` id or request id
        :param update_listings: update confirmation details if its type is listing.
            Needed to map confirmation to listing
        :return: `Confirmation`
        :raises KeyError: when unable to find confirmation by key
        :raises EResultError: for ordinary reasons
        """

        confs = await self.get_confirmations(update_listings=update_listings)
        # not well performant but anyway
        conf = next(filter(lambda c: c.creator_id == key or c.listing_item_ident_code == key, confs), None)
        if conf is None:
            raise KeyError(f"Unable to find confirmation for {key} ident/trade/listing id")

        return conf

    async def allow_all_confirmations(self) -> list[Confirmation]:
        """
        Fetch all confirmations and allow them with single request to `Steam`.

        :return: list of allowed confirmations
        :raises EResultError: for ordinary reasons
        """

        confs = await self.get_confirmations()
        confs and await self.allow_multiple_confirmations(confs)
        return confs

    def allow_confirmation(self, conf: Confirmation) -> CORO[None]:
        """Shorthand for `send_confirmation(conf, 'allow')`."""

        return self.send_confirmation(conf, "allow")

    async def send_confirmation(self, conf: Confirmation, tag: CONF_OP_TAGS):
        """
        Perform confirmation action.

        :param conf: `Confirmation` that you want to proceed
        :param tag: string literal of confirmation tag. Can be 'allow' or 'cancel'
        """

        params = await self._create_confirmation_params(tag)
        params |= {"op": tag, "cid": conf.id, "ck": conf.nonce}
        r = await self.session.get(CONF_URL / "ajaxop", params=params)
        rj = await r.json()

        success = EResult(rj.get("success"))
        if success is not EResult.OK:
            raise EResultError(rj.get("message", "Failed to perform confirmation action"), success, rj)

    def allow_multiple_confirmations(self, confs: list[Confirmation]) -> CORO[None]:
        """Shorthand for `send_multiple_confirmations(conf, 'allow')`."""

        return self.send_multiple_confirmations(confs, "allow")

    async def send_multiple_confirmations(self, confs: list[Confirmation], tag: CONF_OP_TAGS):
        """
        Perform confirmation action for multiple confs with single request to Steam.

        :param confs: list of `Confirmation` that you wand to proceed
        :param tag: string literal of confirmation tag. Can be 'allow' or 'cancel'
        """

        data = await self._create_confirmation_params(tag)
        data |= {"op": tag, "cid[]": [conf.id for conf in confs], "ck[]": [conf.nonce for conf in confs]}
        r = await self.session.post(CONF_URL / "multiajaxop", data=data)
        rj: dict = await r.json()
        success = EResult(rj.get("success"))
        if success is not EResult.OK:
            raise EResultError(rj.get("message", "Failed to perform action for multiple confirmations"), success, rj)

    async def get_confirmations(self, *, update_listings=True) -> list[Confirmation]:
        """
        Fetch all confirmations.

        :param update_listings: update confirmation details if its type is listing.
            Needed to map confirmation to listing.
        :return: list of `Confirmation`
        :raises EResultError: for ordinary reasons
        :raises SessionExpired:
        """

        tag = "getlist"
        params = await self._create_confirmation_params(tag)
        r = await self.session.get(CONF_URL / tag, params=params)
        rj: dict = await r.json()
        success = EResult(rj.get("success"))
        if success is not EResult.OK:
            # https://github.com/DoctorMcKay/node-steamcommunity/blob/1067d4572ee9d467e8f686951901c51028c5c995/components/confirmations.js#L35
            if rj.get("needauth"):
                raise SessionExpired

            raise EResultError(rj.get("message", "Failed to fetch confirmations"), success, rj)

        confs = []
        if "conf" in rj:
            for conf_data in rj["conf"]:
                conf = Confirmation(
                    id=int(conf_data["id"]),
                    nonce=conf_data["nonce"],
                    creator_id=int(conf_data["creator_id"]),
                    creation_time=datetime.fromtimestamp(conf_data["creation_time"]),
                    type=ConfirmationType.get(conf_data["type"]),
                    icon=conf_data["icon"],
                    multi=conf_data["multi"],
                    headline=conf_data["headline"],
                    summary=conf_data["summary"][0],
                    warn=conf_data["warn"],
                )
                if update_listings and conf.type is ConfirmationType.LISTING:
                    # get details so we can find confirmation for listing
                    await self.update_confirmation_with_details(conf)
                confs.append(conf)

        return confs

    async def _create_confirmation_params(self, tag: str) -> dict:
        conf_key, ts = await self._gen_confirmation_key(tag=tag)
        return {
            "p": self.device_id,
            "a": self.steam_id,
            "k": conf_key,
            "t": ts,
            "m": "android",
            "tag": tag,
        }

    async def get_confirmation_details(self, obj: Confirmation | int) -> dict[str, ...]:
        """
        Fetch confirmation details from `Steam`.

        :param obj: `Confirmation` or confirmation id
        :return: dict with details
        :raises EResultError: for ordinary reasons
        """

        if isinstance(obj, Confirmation):
            conf_id = obj.id
        else:
            conf_id = obj

        params = await self._create_confirmation_params(f"details{conf_id}")
        r = await self.session.get(CONF_URL / f"details/{conf_id}", params=params)
        rj = await r.json()
        success = EResult(rj.get("success"))
        if success is not EResult.OK:
            raise EResultError(rj.get("message", "Failed to fetch confirmation details"), success, rj)

        return loads(ITEM_INFO_RE.search(rj["html"])["item_info"])  # TODO TypedDict

    async def update_confirmation_with_details(self, conf: Confirmation):
        """Get confirmation details and update passed `Confirmation` with them."""

        conf.details = await self.get_confirmation_details(conf)



from re import compile
from typing import overload, Literal

from yarl import URL
from aiohttp import ClientResponseError

from ..constants import STEAM_URL, EResult, T_PARAMS, T_HEADERS
from ..exceptions import EResultError, SteamError, SessionExpired
from .confirmation import ConfirmationMixin

API_KEY_RE = compile(r"<p>Key: (?P<api_key>[0-9A-F]+)</p>")
STEAM_GUARD_REQ_CHECK_RE = compile(r"Your account requires (<a [^>]+>)?Steam Guard Mobile Authenticator")


class SteamWebApiMixin(ConfirmationMixin):
    """
    Contain methods related to `Steam Web API`.
    Depends on `ConfirmationMixin`.

    .. seealso:: https://steamapi.xpaw.me
    """

    __slots__ = ()

    # required instance attributes
    _api_key: str | None  # optional

    @overload
    async def call_web_api(
        self,
        url: str | URL,
        *,
        params: T_PARAMS = ...,
        use_api_key: bool = ...,
        headers: T_HEADERS = ...,
    ) -> dict[str, ...]:
        ...

    @overload
    async def call_web_api(
        self,
        url: str | URL,
        *,
        data=...,
        use_api_key: bool = ...,
        headers: T_HEADERS = ...,
        method: Literal["POST"],
    ) -> dict[str, ...]:
        ...

    @overload
    async def call_web_api(
        self,
        url: str | URL,
        *,
        json=...,
        use_api_key: bool = ...,
        headers: T_HEADERS = ...,
        method: Literal["POST"],
    ) -> dict[str, ...]:
        ...

    # https://github.com/DoctorMcKay/node-steam-tradeoffer-manager/blob/7d27ae16642ad810a44d1aed7837872b92392daf/lib/webapi.js#L7
    async def call_web_api(
        self,
        url: str | URL,
        *,
        params={},
        data=None,
        json=None,
        use_api_key=False,
        headers: T_HEADERS = None,
        method: Literal["GET", "POST"] = "GET",
    ) -> dict[str, ...]:
        """
        Make request to a `Steam Web API`

        :param url:
        :param params: params to pass with url
        :param data: form data to send with request
        :param json: json data to send with request
        :param use_api_key: force to use `Steam Web API` key instead of access token
        :param headers:
        :param method: http request method
        :return: json-loaded data
        :raises SessionExpired:
        :raises SteamError:
        """

        params = params.copy()
        if use_api_key:
            if not self._api_key:
                raise AttributeError("You must set an `_api_key` before use this method with `use_api_key=True`")
            params["key"] = self._api_key
        else:
            params["access_token"] = self.access_token

        try:
            r = await self.session.request(method, url, params=params, data=data, json=json, headers=headers)
        except ClientResponseError as e:
            if e.status == 403:
                if not use_api_key and self.is_access_token_expired:
                    raise SessionExpired from e
                else:
                    raise SteamError(f"{'Steam Web API key' if use_api_key else 'Access token'} is invalid") from e
            else:
                raise e

        # https://github.com/DoctorMcKay/node-steam-tradeoffer-manager/blob/7d27ae16642ad810a44d1aed7837872b92392daf/lib/webapi.js#L56
        result = EResult(int(r.headers["X-Eresult"]))
        if r.content.total_bytes > 0:
            rj: dict = await r.json()
            if len(rj) > 1 or len(rj.get("response", ())) > 0:
                return rj

        elif result is not EResult.OK:
            raise EResultError(f"Failed to make {method} request to '{url}'", result)
        else:
            raise SteamError("Invalid response")

    async def get_api_key(self) -> str:
        """
        Fetch `Steam Web API` key, cache it and return.

        :raises SteamError:
        """

        # https://github.com/DoctorMcKay/node-steamcommunity/blob/b58745c8b74963eae808d33e558dbba6840c7053/components/webapi.js#L18
        # force english
        r = await self.session.get(STEAM_URL.COMMUNITY / "dev/apikey", params={"l": "english"}, allow_redirects=False)
        rt = await r.text()

        if "You must have a validated email address to create a Steam Web API key" in rt:
            raise SteamError("Validated email address required to create a Steam Web API key")
        elif STEAM_GUARD_REQ_CHECK_RE.search(rt):
            # for practically impossible case when `shared_secret` is "" and mobile authenticator disabled
            raise SteamError("Steam Guard Mobile Authenticator is required")
        elif "<h2>Access Denied</h2>" in rt:
            raise SteamError("Access to Steam Web Api page is denied")

        search = API_KEY_RE.search(rt)
        if not search:
            raise SteamError("Failed to get Steam Web API key", rt)

        self._api_key = search["api_key"]
        return self._api_key

    async def revoke_api_key(self):
        """Revoke old `Steam Web API` key"""

        data = {
            "sessionid": self.session_id,
            "Revoke": "Revoke My Steam Web API Key",  # whatever
        }
        await self.session.post(STEAM_URL.COMMUNITY / "dev/revokekey", data=data, allow_redirects=False)
        self._api_key = None

    async def register_new_api_key(self, domain: str) -> str:
        """
        Request registration of a new `Steam Web API` key, confirm, cache it and return.

        :param domain: on which domain api key will be registered
        :return: `Steam Web API` key
        :raises EResultError: for ordinary reasons
        """

        # https://github.com/DoctorMcKay/node-steamcommunity/blob/b58745c8b74963eae808d33e558dbba6840c7053/components/webapi.js#L78

        await self.revoke_api_key()  # revoke old one as website do

        data = {
            "domain": domain,
            "request_id": 0,
            "sessionid": self.session_id,
            "agreeToTerms": "true",  # or boolean True?
        }
        r = await self.session.post(STEAM_URL.COMMUNITY / "dev/requestkey", data=data)
        rj: dict[str, str | int] = await r.json()
        success = EResult(rj.get("success"))

        if success is EResult.PENDING and rj.get("requires_confirmation"):
            await self.confirm_api_key_request(rj["request_id"])
            r = await self.session.post(r.url, data=data)  # repeat
            rj: dict[str, str | int] = await r.json()
            success = EResult(rj.get("success"))

        if success is not EResult.OK or not rj["api_key"]:
            raise EResultError(rj.get("message", "Failed to register Steam Web API Key"), success, rj)

        self._api_key = rj["api_key"]
        return self._api_key



"""User agents service extension."""

from collections import UserList
from random import choice

from yarl import URL
from aiohttp import ClientSession

__all__ = ("UserAgentsService", "SOURCE_URL")

SOURCE_URL = URL("https://raw.githubusercontent.com/somespecialone/random-user-agent/refs/heads/master/ua.txt")


class UserAgentsService(UserList[str]):
    """
    List-like class of user agents responsible for loading and getting random user agents.

    .. seealso:: https://github.com/somespecialone/random-user-agent
    """

    @property
    def agents(self) -> list[str]:
        return self.data

    async def load(self, source_url=SOURCE_URL):
        """Load random user agents from `source url`"""

        async with ClientSession(raise_for_status=True) as sess:
            r = await sess.get(source_url)
            agents: list[str] = (await r.text()).splitlines()

        self.data = agents

    def get_random(self) -> str:
        return choice(self.agents)



from urllib.parse import quote
from re import search as re_search
from json import loads as jloads, dumps as jdumps
from pathlib import Path

from yarl import URL

from ..exceptions import EResultError
from ..constants import STEAM_URL, EResult
from ..typed import ProfileData, PrivacySettingsOptions, CommentPrivacySettingsOptions, AvatarUploadData
from ..utils import to_int_boolean
from .login import LoginMixin


class ProfileMixin(LoginMixin):
    """
    Profile attributes and data related methods.
    Depends on `LoginMixin`.
    """

    __slots__ = ()

    # required instance attributes
    trade_token: str | None

    @property
    def trade_url(self) -> URL | None:
        if self.trade_token:
            return STEAM_URL.TRADE / "new/" % {"partner": self.account_id, "token": self.trade_token}

    @property
    def profile_url(self) -> URL:
        return STEAM_URL.COMMUNITY / f"profiles/{self.steam_id}"

    async def get_profile_url_alias(self) -> URL:
        """Get profile url alias like `https://steamcommunity.com/id/<ALIAS>`"""

        r = await self.session.get(STEAM_URL.COMMUNITY / "my", allow_redirects=False)
        return URL(r.headers["Location"])

    async def register_new_trade_url(self) -> URL:
        """Register new trade url. Cache token."""

        r = await self.session.post(
            self.profile_url / "tradeoffers/newtradeurl",
            data={"sessionid": self.session_id},
        )
        token: str = await r.json()

        self.trade_token = quote(token, safe="~()*!.'")  # https://stackoverflow.com/a/72449666/19419998
        return self.trade_url

    async def get_trade_token(self) -> str | None:
        """Fetch trade token from `Steam`, cache it and return"""

        r = await self.session.get(self.profile_url / "tradeoffers/privacy")
        rt = await r.text()

        search = re_search(r"\d+&token=(?P<token>.+)\" readonly", rt)
        self.trade_token = search["token"] if search else None
        return self.trade_token

    async def get_profile_data(self, profile_alias: URL = None) -> ProfileData:
        """Fetch profile data, including general properties and privacy settings"""

        if profile_alias is None:
            profile_alias = await self.get_profile_url_alias()
        r = await self.session.get(profile_alias / "edit/info")
        rt = await r.text()

        return jloads(re_search(r"data-profile-edit=\"(.+)\" data-profile-badges", rt)[1].replace("&quot;", '"'))

    async def edit_profile(
        self,
        *,
        persona_name: str = None,
        real_name: str = None,
        summary: str = None,
        country: str = None,
        state: str = None,
        city: str = None,
        custom_url: str = None,
        hide_profile_award: bool = None,
    ):
        """
        Edit profile general data

        :param persona_name: nickname
        :param real_name: real name of the user
        :param summary: profile summary
        :param country:
        :param state:
        :param city:
        :param custom_url: custom url `ALIAS` (`https://steamcommunity.com/id/<ALIAS>`)
        :param hide_profile_award:
        :raises EResultError: for ordinary reasons
        """

        args = [persona_name, real_name, summary, country, state, city, custom_url, hide_profile_award]
        if all(map(lambda x: x is None, args)):
            raise ValueError("You need to pass at least one argument value!")

        profile_alias = await self.get_profile_url_alias()
        profile_data = await self.get_profile_data(profile_alias)

        # https://github.com/DoctorMcKay/node-steamcommunity/blob/1067d4572ee9d467e8f686951901c51028c5c995/components/profile.js#L56
        data = {
            "sessionID": self.session_id,
            "type": "profileSave",
            "hide_profile_awards": "0",
            "json": 1,
            "weblink_1_title": "",
            "weblink_1_url": "",
            "weblink_2_title": "",
            "weblink_2_url": "",
            "weblink_3_title": "",
            "weblink_3_url": "",
            # attrs below
            "personaName": persona_name if persona_name is not None else profile_data["strPersonaName"],
            "real_name": real_name if real_name is not None else profile_data["strRealName"],
            "hide_profile_award": to_int_boolean(hide_profile_award)
            if hide_profile_award is not None
            else profile_data["ProfilePreferences"]["hide_profile_awards"],
            "summary": summary if summary is not None else profile_data["strSummary"],
            "country": country if country is not None else profile_data["LocationData"]["locCountryCode"],
            "state": state if state is not None else profile_data["LocationData"]["locStateCode"],
            "city": city if city is not None else profile_data["LocationData"]["locCity"],
            "customURL": custom_url if custom_url is not None else profile_data["strCustomURL"],
        }
        headers = {"Referer": str(profile_alias / "edit/settings")}

        r = await self.session.post(profile_alias / "edit/", data=data, headers=headers)
        rj = await r.json()
        success = EResult(rj.get("success"))
        if success is not EResult.OK:
            raise EResultError(rj.get("message", "Failed to edit profile"), success, rj)

    async def edit_privacy_settings(
        self,
        *,
        friends_list: PrivacySettingsOptions = None,
        inventory: PrivacySettingsOptions = None,
        inventory_gifts: bool = None,
        owned_games: PrivacySettingsOptions = None,
        playtime: bool = None,
        profile: PrivacySettingsOptions = None,
        comment_permission: CommentPrivacySettingsOptions = None,
    ):
        """
        Edit profile privacy settings. Use values from below:

        Privacy levels:
            * 1 - private
            * 2 - friends only
            * 3 - public

        Comment permission:
            * 2 - private
            * 0 - friends only
            * 1 - public

        :param friends_list: desired privacy level required to view your friends list
        :param inventory: your desired `Steam` inventory privacy state
        :param inventory_gifts: to keep your `Steam` gift inventory private
        :param owned_games: your desired privacy level required to view games you own and what game you're
            currently playing
        :param playtime: to keep your game playtime private
        :param profile: your desired profile privacy state
        :param comment_permission: your desired profile comments privacy state
        :raises EResultError: for ordinary reasons
        """

        args = [
            friends_list,
            inventory,
            inventory_gifts,
            owned_games,
            playtime,
            profile,
            comment_permission,
        ]
        if all(map(lambda x: x is None, args)):
            raise ValueError("You need to pass at least one argument value!")

        profile_alias = await self.get_profile_url_alias()
        profile_data = await self.get_profile_data(profile_alias)

        privacy_settings = profile_data["Privacy"]["PrivacySettings"]
        comment_permission = profile_data["Privacy"]["eCommentPermission"]

        if friends_list is not None:
            privacy_settings["PrivacyFriendsList"] = friends_list
        if inventory is not None:
            privacy_settings["PrivacyInventory"] = inventory
        if inventory_gifts is not None:
            privacy_settings["PrivacyInventoryGifts"] = 3 if inventory_gifts else 1
        if owned_games is not None:
            privacy_settings["PrivacyOwnedGames"] = owned_games
        if playtime is not None:
            privacy_settings["PrivacyPlaytime"] = 3 if playtime else 1
        if profile is not None:
            privacy_settings["PrivacyProfile"] = profile
        if comment_permission is not None:
            comment_permission = comment_permission

        data = {
            "sessionid": self.session_id,
            "Privacy": jdumps(privacy_settings),
            "eCommentPermission": comment_permission,
        }
        headers = {"Referer": str(profile_alias / "edit/settings")}

        r = await self.session.post(profile_alias / "ajaxsetprivacy/", data=data, headers=headers)
        rj = await r.json()
        success = EResult(rj.get("success"))
        if success is not EResult.OK:
            raise EResultError(rj.get("message", "Failed to edit profile privacy settings"), success, rj)

    async def upload_avatar(self, source: Path | URL | bytes) -> AvatarUploadData:
        """
        Replaces your current avatar image with a new one.
        `Source` can be a `pathlib.Path` pointing to a file, a `yarl.URL` as an address to image on a remote source,
        or a binary bytes buffer.
        """

        if isinstance(source, Path):
            with source.open("rb") as f:
                source = f.read()

        elif isinstance(source, URL):
            async with self.session.get(source) as r:
                source = await r.content.read()

        data = {
            "type": "player_avatar_image",
            "sId": str(self.steam_id),
            "sessionid": self.session_id,
            "doSub": "1",
            "json": "1",
            "avatar": source,
        }

        r = await self.session.post(STEAM_URL.COMMUNITY / "actions/FileUploader", data=data)
        rj = await r.json()
        success = EResult(rj.get("success"))
        if success is not EResult.OK:
            raise EResultError(rj.get("message", "Failed to upload avatar"), success, rj)

        return rj



from dataclasses import dataclass, field
from typing import NamedTuple
from datetime import datetime

from yarl import URL

from .constants import (
    STEAM_URL,
    TRADABLE_AFTER_DATE_FORMAT,
    App,
    AppContext,
    Currency,
    ConfirmationType,
    MarketHistoryEventType,
    MarketListingStatus,
    TradeOfferStatus,
)
from .utils import create_ident_code, account_id_to_steam_id, make_inspect_url


class ItemAction(NamedTuple):
    link: str
    name: str


class ItemDescriptionEntry(NamedTuple):
    value: str
    color: str | None  # hexadecimal


class ItemTag(NamedTuple):
    category: str
    internal_name: str
    localized_category_name: str
    localized_tag_name: str
    color: str | None  # hexadecimal


@dataclass(eq=False, slots=True, frozen=True, kw_only=True)
class ItemDescription:
    """
    `EconItem` description representation.
    `id` or `ident_code` field is guaranteed unique within whole Steam Economy.
    """

    id: str = field(init=False, default="")  # optimization 🚀
    """Unique identifier of the `ItemDescription` within `Steam Economy`"""

    class_id: int
    instance_id: int

    d_id: int | None = field(init=False, default=None)  # optional CSGO inspect id

    app: App

    name: str
    market_name: str
    market_hash_name: str

    type: str | None = None

    name_color: str | None = None  # hexadecimal
    background_color: str | None = None

    icon: str
    icon_large: str | None = None

    actions: tuple[ItemAction, ...] = ()
    market_actions: tuple[ItemAction, ...] = ()
    owner_actions: tuple[ItemAction, ...] = ()
    tags: tuple[ItemTag, ...] = ()
    descriptions: tuple[ItemDescriptionEntry, ...] = ()
    owner_descriptions: tuple[ItemDescriptionEntry, ...] = ()

    fraud_warnings: tuple[str, ...] = ()

    commodity: bool  # item use buy orders on market
    tradable: bool  # item can be traded
    marketable: bool
    # days for which the item will be untradable after being sold on the market.
    market_tradable_restriction: int | None = None
    market_buy_country_restriction: str | None = None
    market_fee_app: int | None = None
    market_marketable_restriction: int | None = None  # same as `market_tradable_restriction` but for market

    def __post_init__(self):
        self._set_ident_code()
        self._set_d_id()

    def _set_ident_code(self):
        object.__setattr__(self, "id", create_ident_code(self.instance_id, self.class_id, self.app.value))

    def _set_d_id(self):
        if self.app is App.CS2:
            if (i_action := next(filter(lambda a: "Inspect" in a.name, self.actions), None)) is not None:
                object.__setattr__(self, "d_id", int(i_action.link.split("%D")[1]))

    @property
    def ident_code(self) -> str:
        """Alias for `id`"""
        return self.id

    @property
    def icon_url(self) -> URL:
        return STEAM_URL.STATIC / f"economy/image/{self.icon}/96fx96f"

    @property
    def icon_large_url(self) -> URL | None:
        return (STEAM_URL.STATIC / f"economy/image/{self.icon_large}/330x192") if self.icon_large is not None else None

    @property
    def market_url(self) -> URL:
        return STEAM_URL.MARKET / f"listings/{self.app.value}/{self.market_hash_name}"

    def __eq__(self, other):
        if isinstance(other, ItemDescription):
            return self.id == other.id
        return False

    def __hash__(self):
        return hash(self.id)


@dataclass(eq=False, slots=True, kw_only=True)
class EconItem:
    """
    Represents unique copy of a `Steam Economy` item, ala `Asset`.
    `id` or `ident_code` field is guaranteed unique within whole Steam Economy.
    """

    id: str = field(init=False, default="")  # optimization 🚀
    """Unique identifier of the `EconItem` within `Steam Economy`"""

    asset_id: int  # The item's unique ID within its app+context
    owner_id: int

    app_context: AppContext

    amount: int  # if stackable

    description: ItemDescription
    tradable_after: datetime | None = field(init=False, default=None)

    def __post_init__(self):
        self._set_ident_code()
        self._set_tradable_after()

    def _set_ident_code(self):
        self.id = create_ident_code(self.asset_id, self.app_context.context, self.app_context.app.value)

    def _set_tradable_after(self):
        if self.description.market_tradable_restriction:
            sep = "Tradable After "
            t_a_descr = next(filter(lambda d: sep in d.value, self.description.owner_descriptions or ()), None)
            if t_a_descr is not None:
                self.tradable_after = datetime.strptime(t_a_descr.value, TRADABLE_AFTER_DATE_FORMAT)

    @property
    def ident_code(self) -> str:
        """Alias for `id`"""
        return self.id

    @property
    def inspect_url(self) -> str | None:
        if self.description.d_id:
            return make_inspect_url(owner_id=self.owner_id, asset_id=self.asset_id, d_id=self.description.d_id)

    def __eq__(self, other):
        if isinstance(other, EconItem):
            return self.id == other.id
        return False

    def __hash__(self):
        return hash(self.id)


# https://github.com/DoctorMcKay/node-steamcommunity/wiki/CConfirmation
@dataclass(eq=False, slots=True)
class Confirmation:
    id: int
    nonce: str  # conf key
    creator_id: int  # trade/listing id
    creation_time: datetime

    type: ConfirmationType

    icon: str
    multi: bool  # ?
    headline: str
    summary: str
    warn: str | None  # ?

    details: dict[str, ...] | None = None  # TODO need typing

    @property
    def listing_item_ident_code(self) -> str | None:
        """`MarketListingItem` ident code if `details` is present"""
        if self.details is not None:
            return create_ident_code(self.details["id"], self.details["contextid"], self.details["appid"])


class Notifications(NamedTuple):
    trades: int  # 1
    game_turns: int  # 2
    moderator_messages: int  # 3
    comments: int  # 4
    items: int  # 5
    invites: int  # 6
    # 7 missing
    gifts: int  # 8
    chats: int  # 9
    help_request_replies: int  # 10
    account_alerts: int  # 11


@dataclass(eq=False, slots=True, kw_only=True)
class MarketListingItem(EconItem):
    """Presented only on active listing."""

    market_id: int  # listing id

    unowned_id: int | None
    unowned_context_id: int | None

    amount: int = 1  # item on listing always have amount eq 1
    owner_id: int = 0

    @property
    def inspect_url(self) -> str | None:
        if self.description.d_id:
            return make_inspect_url(market_id=self.market_id, asset_id=self.asset_id, d_id=self.description.d_id)


@dataclass(eq=False, slots=True)
class BaseOrder:
    id: int  # listing/buy order id

    price: int

    def __hash__(self):
        return self.id


@dataclass(eq=False, slots=True)
class MyMarketListing(BaseOrder):
    lister_steam_id: int
    time_created: datetime

    item: MarketListingItem

    status: MarketListingStatus
    active: bool  # ?

    # fields that can be useful
    item_expired: int
    cancel_reason: int
    time_finish_hold: int

    @property
    def listing_id(self) -> int:
        return self.id


@dataclass(eq=False, slots=True)
class BuyOrder(BaseOrder):
    item_description: ItemDescription

    quantity: int
    quantity_remaining: int

    @property
    def buy_order_id(self) -> int:
        return self.id


@dataclass(eq=False, slots=True, kw_only=True)
class MarketHistoryListingItem(MarketListingItem):
    market_id: None = None

    # purchase fields
    new_asset_id: int | None = None
    new_context_id: int | None = None
    rollback_new_asset_id: int | None = None
    rollback_new_context_id: int | None = None

    @property
    def inspect_link(self) -> None:
        """Always `None`, because we can't be sure that asset id has not been changed."""
        return None


@dataclass(eq=False, slots=True, kw_only=True)
class MarketHistoryListing(BaseOrder):
    item: MarketHistoryListingItem

    currency: Currency

    # purchase fields
    purchase_id: int | None = None
    steamid_purchaser: int | None = None
    received_amount: int | None = None
    received_currency: Currency | None = None
    time_sold: datetime | None = None
    paid_amount: int | None = None
    paid_fee: int | None = None
    steam_fee: int | None = None
    publisher_fee: int | None = None

    # unknown fields
    # failed
    # needs_rollback
    # funds_held
    # time_funds_held_until
    # funds_revoked
    # funds_returned

    # listing fields
    price: int | None = None
    fee: int | None = None
    original_price: int | None = None
    cancel_reason: str | None = None

    @property
    def listing_id(self) -> int:
        return self.id


@dataclass(eq=False, slots=True)
class MarketHistoryEvent:
    """
    Event entity in market history. Represents event linked with related listing.
    Note that this is just a snapshot of listing, asset data for event fire moment time
    and `asset id`, `context id` of asset may change already.
    """

    listing: MarketHistoryListing
    time_event: datetime
    type: MarketHistoryEventType


@dataclass(eq=False, slots=True)
class PriceHistoryEntry:
    date: datetime
    price: float  # float from steam
    daily_volume: int


@dataclass(eq=False, slots=True)
class MarketListing(BaseOrder):
    item: MarketListingItem

    currency: Currency  # original currency
    fee: int

    # probably Steam do not return converted values if fee and price is already in requested currency
    converted_currency: Currency | None
    converted_price: int | None
    converted_fee: int | None

    def __post_init__(self):
        if not self.item.market_id:
            self.item.market_id = self.id

    @property
    def listing_id(self) -> int:
        return self.id

    @property
    def total_cost(self) -> int:
        return self.price + self.fee

    @property
    def total_converted_cost(self) -> int | None:
        if self.converted_price is not None and self.converted_fee is not None:
            return self.converted_price + self.converted_fee


@dataclass(eq=False, slots=True, kw_only=True)
class BaseTradeOfferItem(EconItem):
    description: ItemDescription | None  # sometimes data has no description, because of course

    def _set_tradable_after(self):
        if self.description is not None and self.description.market_tradable_restriction:
            # cannot do super()._set_tradable_after() due to super exception
            t_a_descr = next(
                filter(lambda d: "Tradable After " in d.value, self.description.owner_descriptions or ()),
                None,
            )
            if t_a_descr is not None:
                self.tradable_after = datetime.strptime(t_a_descr.value, TRADABLE_AFTER_DATE_FORMAT)

    @property
    def inspect_url(self) -> str | None:
        if self.description is not None and self.description.d_id:  # can't do super().inspect_url due to an error
            return make_inspect_url(owner_id=self.owner_id, asset_id=self.asset_id, d_id=self.description.d_id)


@dataclass(eq=False, slots=True, kw_only=True)
class TradeOfferItem(BaseTradeOfferItem):
    missing: bool
    est_usd: int

    owner_id: int = 0


@dataclass(eq=False, slots=True)
class BaseTradeOffer:
    id: int

    owner_id: int  # steam id64 of entity owner acc (SteamClient)
    partner_id: int  # id32

    status: TradeOfferStatus

    @property
    def trade_offer_id(self) -> int:
        return self.id

    @property
    def partner_id64(self) -> int:
        return account_id_to_steam_id(self.partner_id)

    # TODO to remove in 0.8.0
    @property
    def is_active(self) -> bool:
        from warnings import warn

        warn(
            "`is_active` property of `BaseTradeOffer` is deprecated. Use `active` instead",
            DeprecationWarning,
            stacklevel=2,
        )
        return self.active

    # shorthands
    @property
    def active(self) -> bool:
        return self.status is TradeOfferStatus.ACTIVE

    @property
    def accepted(self):
        return self.status is TradeOfferStatus.ACCEPTED

    @property
    def declined(self):
        return self.status is TradeOfferStatus.DECLINED

    def __hash__(self):
        return self.id


@dataclass(eq=False, slots=True)
class TradeOffer(BaseTradeOffer):
    """Steam Trade Offer entity."""

    is_our_offer: bool

    expiration_time: datetime
    time_created: datetime
    time_updated: datetime

    items_to_give: list[TradeOfferItem]
    items_to_receive: list[TradeOfferItem]

    message: str = ""

    def __post_init__(self):
        for i in self.items_to_give:
            i.owner_id = self.owner_id

        for i in self.items_to_receive:
            i.owner_id = self.partner_id64

    @property
    def sender(self) -> int:
        """Steam id64 of sender."""
        return self.owner_id if self.is_our_offer else self.partner_id64

    @property
    def receiver(self) -> int:
        """Steam id64 of receiver."""
        return self.partner_id64 if self.is_our_offer else self.owner_id


@dataclass(eq=False, slots=True, kw_only=True)
class HistoryTradeOfferItem(BaseTradeOfferItem):
    new_asset_id: int
    new_context_id: int

    owner_id: None = None


@dataclass(eq=False, slots=True)
class HistoryTradeOffer(BaseTradeOffer):
    time_init: datetime

    assets_received: list[HistoryTradeOfferItem]
    assets_given: list[HistoryTradeOfferItem]


class SellOrderTableEntry(NamedTuple):
    price: int
    price_with_fee: int
    quantity: int


class BuyOrderTableEntry(NamedTuple):
    price: int
    quantity: int


class OrderGraphEntry(NamedTuple):
    price: int
    quantity: int
    repr: str


@dataclass(eq=False, slots=True)
class ItemOrdersHistogram:
    sell_order_count: int
    sell_order_price: int
    sell_order_table: list[SellOrderTableEntry]
    buy_order_count: int
    buy_order_price: int
    buy_order_table: list[BuyOrderTableEntry]
    highest_buy_order: int
    lowest_sell_order: int

    # prices in integers (cents)!
    buy_order_graph: list[OrderGraphEntry]
    sell_order_graph: list[OrderGraphEntry]

    graph_max_y: int
    graph_min_x: int  # in cents
    graph_max_x: int  # in cents
    # price_prefix: str | None
    # price_suffix: str | None


@dataclass(eq=False, slots=True)
class MarketSearchItem:
    sell_listings: int
    sell_price: int
    sell_price_text: str
    sale_price_text: str

    app_icon: str
    app_name: str

    description: ItemDescription



from contextlib import suppress
from typing import TypeAlias, overload, AsyncIterator, Literal, Callable, Sequence, Mapping
from datetime import datetime
from re import compile as re_compile

from aiohttp import ClientResponseError

from ..constants import STEAM_URL, App, AppContext, Currency, T_PARAMS, T_HEADERS, EResult
from ..helpers import currency_required
from ..typed import ItemOrdersHistogramData, ItemOrdersActivity, PriceOverview
from ..exceptions import EResultError, SteamError, RateLimitExceeded, ResourceNotModified
from ..utils import create_ident_code, find_item_nameid_in_text, parse_time, format_time, to_int_boolean
from ..models import (
    ItemDescriptionEntry,
    ItemTag,
    ItemDescription,
    EconItem,
    ItemAction,
    MarketListing,
    MarketListingItem,
    ItemOrdersHistogram,
    SellOrderTableEntry,
    BuyOrderTableEntry,
    OrderGraphEntry,
    MarketSearchItem,
)
from .http import SteamHTTPTransportMixin


# steam limit rules
INV_COUNT = 5000
LISTING_COUNT = 10

INVENTORY_URL = STEAM_URL.COMMUNITY / "inventory"
SEARCH_URL = STEAM_URL.MARKET / "search"
SEARCH_RENDER_URL = SEARCH_URL / "render/"

# listings, total count, last modified
T_MARKET_ITEM_LISTINGS_DATA: TypeAlias = tuple[list[MarketListing], int, datetime]
INV_ITEM_DATA: TypeAlias = tuple[list[EconItem], int, int | None]  # items, total count, last asset id for pagination

ITEM_ORDER_HIST_PRICE_RE = re_compile(r"[^\d\s]*([\d,]+(?:\.\d+)?)[^\d\s]*")  # Author: ChatGPT

T_SHARED_DESCRIPTIONS: TypeAlias = dict[str, ItemDescription]  # ident code : descr

T_SORT_COLUMN: TypeAlias = Literal["price", "name", "quantity", "popular", "default"]
T_SORT_DIR: TypeAlias = Literal["asc", "desc"]


class SteamCommunityPublicMixin(SteamHTTPTransportMixin):
    """
    Contains methods that do not require authentication.
    Depends on `SteamHTTPTransportMixin`.
    """

    __slots__ = ()

    # required instance attributes
    currency: Currency | None
    country: str

    async def get_user_inventory(
        self,
        steam_id: int,
        app_context: AppContext,
        *,
        last_assetid: int = None,
        count=INV_COUNT,
        params: T_PARAMS = {},
        headers: T_HEADERS = {},
        _item_descriptions_map: T_SHARED_DESCRIPTIONS = None,
    ) -> INV_ITEM_DATA:
        """
        Fetches inventory of user.

        .. note::
            * You can paginate by yourself passing `last_assetid` arg
            * `count` arg value that less than 2000 lead to responses with strange amount of assets

        :param steam_id: steamid64 of user
        :param app_context: `Steam` app+context
        :param last_assetid:
        :param count: page size
        :param params: extra params to pass to url
        :param headers: extra headers to send with request
        :return: list of `EconItem`, total count of items in inventory, last asset id of the list
        :raises SteamError: if inventory is private
        :raises EResultError: for ordinary reasons
        :raises RateLimitExceeded: when you hit rate limit
        """

        inv_url = INVENTORY_URL / f"{steam_id}/"
        params = {"l": self.language, "count": count, **params}
        if last_assetid:
            params["last_assetid"] = last_assetid
        headers = {"Referer": str(inv_url), **headers}

        try:
            r = await self.session.get(
                inv_url / f"{app_context.app.value}/{app_context.context}",
                params=params,
                headers=headers,
            )
        except ClientResponseError as e:
            if e.status == 403:
                # https://github.com/DoctorMcKay/node-steamcommunity/blob/master/components/users.js#L603
                raise SteamError("Steam user inventory is private") from e
            elif e.status == 429:  # never faced this, but let it be
                raise RateLimitExceeded("You have been rate limited, rest for a while!") from e
            else:
                raise e

        rj: dict[str, list[dict] | int] = await r.json()
        success = EResult(rj.get("success"))
        if success is not EResult.OK:
            raise EResultError(rj.get("message", "Failed to fetch inventory"), success, rj)

        total_count = rj["total_inventory_count"]
        last_assetid_return = int(rj["last_assetid"]) if "last_assetid" in rj else None

        if "descriptions" not in rj:  # count<=101, last_assetid=None and we got there
            return [], total_count, last_assetid_return

        if _item_descriptions_map is None:
            _item_descriptions_map = {}

        items = self._parse_inventory(rj, steam_id, _item_descriptions_map)

        return items, total_count, last_assetid_return

    @classmethod
    def _parse_inventory(
        cls,
        data: dict[str, list[dict]],
        steam_id: int,
        descrs_map: dict[str, ItemDescription],
    ) -> list[EconItem]:
        for d_data in data["descriptions"]:
            key = create_ident_code(d_data["instanceid"], d_data["classid"], d_data["appid"])
            if key not in descrs_map:
                descrs_map[key] = cls._create_item_descr(d_data)

        return [
            EconItem(
                asset_id=int(a_data["assetid"]),
                owner_id=steam_id,
                amount=int(a_data["amount"]),
                description=descrs_map[create_ident_code(a_data["instanceid"], a_data["classid"], a_data["appid"])],
                app_context=AppContext((App(int(a_data["appid"])), int(a_data["contextid"]))),
            )
            for a_data in data["assets"]
        ]

    @staticmethod
    def _create_item_actions(actions: list[dict]) -> tuple[ItemAction, ...]:
        return tuple(ItemAction(a_data["link"], a_data["name"]) for a_data in actions)

    @staticmethod
    def _create_item_tags(tags: list[dict]) -> tuple[ItemTag, ...]:
        return tuple(
            ItemTag(
                t_data["category"],
                t_data["internal_name"],
                t_data["localized_category_name"],
                t_data["localized_tag_name"],
                t_data.get("color"),
            )
            for t_data in tags
        )

    @staticmethod
    def _create_item_descr_entries(descriptions: list[dict]) -> tuple[ItemDescriptionEntry, ...]:
        return tuple(
            ItemDescriptionEntry(de_data["value"], de_data.get("color"))
            for de_data in descriptions
            if de_data["value"] != " "  # ha, surprise!
        )

    @classmethod
    def _create_item_descr(cls, data: dict) -> ItemDescription:
        return ItemDescription(
            class_id=int(data["classid"]),
            instance_id=int(data["instanceid"]),
            app=App(data["appid"]),
            name=data["name"],
            market_name=data["market_name"],
            market_hash_name=data["market_hash_name"],
            name_color=data.get("name_color") or None,
            background_color=data.get("name_color") or None,
            type=data["type"] or None,
            icon=data["icon_url"],
            icon_large=data.get("icon_url_large"),
            commodity=bool(data["commodity"]),
            tradable=bool(data["tradable"]),
            marketable=bool(data.get("marketable", True)),  # True in case of missing in data from market search page
            market_tradable_restriction=data.get("market_tradable_restriction"),
            market_buy_country_restriction=data.get("market_buy_country_restriction"),
            market_fee_app=data.get("market_fee_app"),
            market_marketable_restriction=data.get("market_marketable_restriction"),
            actions=cls._create_item_actions(data["actions"]) if "actions" in data else (),
            market_actions=cls._create_item_actions(data["market_actions"]) if "market_actions" in data else (),
            owner_actions=cls._create_item_actions(data["owner_actions"]) if "owner_actions" in data else (),
            tags=cls._create_item_tags(data["tags"]) if "tags" in data else (),
            descriptions=cls._create_item_descr_entries(data["descriptions"]) if "descriptions" in data else (),
            owner_descriptions=cls._create_item_descr_entries(data["owner_descriptions"])
            if "owner_descriptions" in data
            else (),
            fraud_warnings=tuple(data.get("fraudwarnings", ())),
        )

    async def user_inventory(
        self,
        steam_id: int,
        app_context: AppContext,
        *,
        last_assetid: int = None,
        count=INV_COUNT,
        params: T_PARAMS = {},
        headers: T_HEADERS = {},
        _item_descriptions_map: T_SHARED_DESCRIPTIONS = None,
    ) -> AsyncIterator[INV_ITEM_DATA]:
        """
        Fetches inventory of user. Return async iterator to paginate over inventory pages.

        .. note:: `count` arg value that less than 2000 lead to responses with strange amount of assets

        :param steam_id: steamid64 of user
        :param app_context: `Steam` app+context
        :param last_assetid:
        :param count: page size
        :param params: extra params to pass to url
        :param headers: extra headers to send with request
        :return: `AsyncIterator` that yields list of `EconItem`, total count of items in inventory, last asset id of the list
        :raises EResultError: for ordinary reasons
        :raises RateLimitExceeded: when you hit rate limit
        :raises SteamError: if inventory is private
        """

        if _item_descriptions_map is None:  # shared descriptions instances across calls
            _item_descriptions_map = {}

        more_items = True
        while more_items:
            # browser do first request with count=75, receiving data with `last_assetid` only
            # avoid excess destructuring
            inventory_data = await self.get_user_inventory(
                steam_id,
                app_context,
                last_assetid=last_assetid,
                count=count,
                params=params,
                headers=headers,
                _item_descriptions_map=_item_descriptions_map,
            )
            last_assetid = inventory_data[2]
            # let's assume that field "last_assetid" always present with "more_items" so we can depend on it
            more_items = bool(last_assetid)

            yield inventory_data

    @overload
    async def get_user_inventory_item(
        self,
        steam_id: int,
        app_context: AppContext,
        obj: int = ...,
        *,
        params: T_PARAMS = ...,
        headers: T_HEADERS = ...,
        **item_attrs,
    ) -> EconItem | None:
        ...

    @overload
    async def get_user_inventory_item(
        self,
        steam_id: int,
        app_context: AppContext,
        obj: Callable[[EconItem], bool],
        *,
        params: T_PARAMS = ...,
        headers: T_HEADERS = ...,
    ) -> EconItem | None:
        ...

    # TODO start_assetid param to find specified asset id
    async def get_user_inventory_item(
        self,
        steam_id: int,
        app_context: AppContext,
        obj: int | Callable[[EconItem], bool] = None,
        *,
        params: T_PARAMS = {},
        headers: T_HEADERS = {},
        _item_descriptions_map: T_SHARED_DESCRIPTIONS = None,
        **item_attrs,
    ) -> EconItem | None:
        """
        Fetch and iterate over inventory item pages of user until find one that satisfies passed arguments.

        :param steam_id: steamid64 of user
        :param app_context: `Steam` app+context
        :param obj: asset id or predicate function
        :param params: extra params to pass to url
        :param headers: extra headers to send with request
        :param item_attrs: additional item attributes and values
        :return: `EconItem` or `None`
        :raises EResultError: for ordinary reasons
        :raises RateLimitExceeded: when you hit rate limit
        :raises SteamError: if inventory is private
        """

        if callable(obj):
            predicate = obj
        else:

            def predicate(i: EconItem):
                if obj is not None and i.asset_id != obj:
                    return False

                for attr, value in item_attrs.items():
                    if getattr(i, attr, None) != value:
                        return False

                return True

        if _item_descriptions_map is None:
            _item_descriptions_map = {}

        async for data in self.user_inventory(
            steam_id,
            app_context,
            params=params,
            headers=headers,
            _item_descriptions_map=_item_descriptions_map,
        ):
            with suppress(StopIteration):
                return next(filter(predicate, data[0]))

    @overload
    async def get_item_orders_histogram(
        self,
        item_nameid: int,
        *,
        if_modified_since: datetime | str = ...,
        params: T_PARAMS = ...,
        headers: T_HEADERS = ...,
    ) -> tuple[ItemOrdersHistogram, datetime]:
        ...

    @overload
    async def get_item_orders_histogram(
        self,
        item_nameid: int,
        *,
        raw: Literal[True] = ...,
        if_modified_since: datetime | str = ...,
        params: T_PARAMS = ...,
        headers: T_HEADERS = ...,
    ) -> tuple[ItemOrdersHistogramData, datetime]:
        ...

    @currency_required
    async def get_item_orders_histogram(
        self,
        item_nameid: int,
        *,
        raw: bool = False,
        if_modified_since: datetime | str = None,
        params: T_PARAMS = {},
        headers: T_HEADERS = {},
    ) -> tuple[ItemOrdersHistogramData | ItemOrdersHistogram, datetime]:
        """
        Do what described in method name.

        .. seealso::
            * https://github.com/Revadike/InternalSteamWebAPI/wiki/Get-Market-Item-Orders-Histogram
            * https://github.com/somespecialone/steam-item-name-ids

        .. note:: This request is rate limited by Steam. It is strongly recommended to use `if_modified_since`

        :param item_nameid: special id of item class. Can be found only on listings page.
        :param raw: if `True`, return `ItemOrdersHistogramData` dict instead of `ItemOrdersHistogram` model
        :param if_modified_since: `If-Modified-Since` header value
        :param params: extra params to pass to url
        :param headers: extra headers to send with request
        :return: `ItemOrdersHistogramData` dict if `raw` is `True` or `ItemOrdersHistogram` model,
            datetime object when resource was last modified
        :raises EResultError: for ordinary reasons
        :raises RateLimitExceeded: when you hit rate limit
        :raises ResourceNotModified: when `if_modified_since` header passed and Steam response with 304 status code
        """

        params = {
            "norender": 1,
            "language": self.language,
            "country": self.country,
            "currency": self.currency,
            "item_nameid": item_nameid,
            **params,
        }
        headers = {**headers}
        if if_modified_since:
            if isinstance(if_modified_since, datetime):
                headers["If-Modified-Since"] = format_time(if_modified_since)
            else:  # str
                headers["If-Modified-Since"] = if_modified_since

        try:
            r = await self.session.get(STEAM_URL.MARKET / "itemordershistogram", params=params, headers=headers)
        except ClientResponseError as e:
            if e.status == 429:
                raise RateLimitExceeded("You have been rate limited, rest for a while!") from e
            else:
                raise e

        last_modified = parse_time(r.headers["Last-Modified"])

        if r.status == 304:  # not modified if header "If-Modified-Since" is provided
            raise ResourceNotModified(last_modified, parse_time(r.headers["Expires"]))

        rj: ItemOrdersHistogramData = await r.json()
        success = EResult(rj.get("success"))
        if success is not EResult.OK:
            raise EResultError(rj.get("message", "Failed to fetch items order histogram"), success, rj)

        if raw:
            return rj, last_modified

        return self._create_item_orders_histogram(rj), last_modified

    @classmethod
    def _create_item_orders_histogram(cls, data: dict) -> ItemOrdersHistogram:
        # model parsing
        return ItemOrdersHistogram(
            sell_order_count=cls._parse_item_order_histogram_count(data["sell_order_count"]),
            sell_order_price=cls._parse_item_order_histogram_price(data["sell_order_price"]),
            sell_order_table=[
                SellOrderTableEntry(
                    cls._parse_item_order_histogram_price(d["price"]),
                    cls._parse_item_order_histogram_price(d["price_with_fee"]),
                    cls._parse_item_order_histogram_count(d["quantity"]),
                )
                for d in data["sell_order_table"]
            ],
            buy_order_count=cls._parse_item_order_histogram_count(data["buy_order_count"]),
            buy_order_price=cls._parse_item_order_histogram_price(data["buy_order_price"]),
            buy_order_table=[
                BuyOrderTableEntry(
                    cls._parse_item_order_histogram_price(d["price"]),
                    cls._parse_item_order_histogram_count(d["quantity"]),
                )
                for d in data["buy_order_table"]
            ],
            highest_buy_order=int(data["highest_buy_order"]),
            lowest_sell_order=int(data["lowest_sell_order"]),
            buy_order_graph=[OrderGraphEntry(int(d[0] * 100), d[1], d[2]) for d in data["buy_order_graph"]],
            sell_order_graph=[OrderGraphEntry(int(d[0] * 100), d[1], d[2]) for d in data["sell_order_graph"]],
            graph_max_y=data["graph_max_y"],
            graph_min_x=int(data["graph_min_x"] * 100),
            graph_max_x=int(data["graph_max_x"] * 100),
        )

    @staticmethod
    def _parse_item_order_histogram_count(text: str) -> int:
        if "." in text:
            count_raw = text.replace(".", "")
        elif "," in text:  # to be sure
            count_raw = text.replace(",", "")
        else:
            count_raw = text

        return int(count_raw)

    @staticmethod
    def _parse_item_order_histogram_price(text: str) -> int:
        raw_price = ITEM_ORDER_HIST_PRICE_RE.search(text).group(1)

        if "." not in raw_price and "," not in raw_price:
            price = int(raw_price) * 100  # add cents
        else:  # 163,46₴ £2.69 $1,000.00
            price = raw_price.replace(",", "").replace(".", "")

        return int(price)

    @currency_required
    async def fetch_item_orders_activity(
        self,
        item_nameid: int,
        *,
        params: T_PARAMS = {},
        headers: T_HEADERS = {},
    ) -> ItemOrdersActivity:
        """
        Do what described in method name.

        .. seealso::
            * https://github.com/Revadike/InternalSteamWebAPI/wiki/Get-Market-Item-Orders-Activity
            * https://github.com/somespecialone/steam-item-name-ids

        :param item_nameid: special id of item class. Can be found only on listings page.
        :param params: extra params to pass to url
        :param headers: extra headers to send with request
        :return: `ItemOrdersActivity` dict
        :raises EResultError:
        """

        params = {
            "norender": 1,
            "language": self.language,
            "country": self.country,
            "currency": self.currency,
            "item_nameid": item_nameid,
            **params,
        }
        r = await self.session.get(STEAM_URL.MARKET / "itemordersactivity", params=params, headers=headers)
        # Can we hit a rate limit there?
        rj: ItemOrdersActivity = await r.json()
        success = EResult(rj.get("success"))
        if success is not EResult.OK:
            raise EResultError(rj.get("message", "Failed to fetch items order activity"), success, rj)

        return rj

    @overload
    async def fetch_price_overview(
        self,
        obj: ItemDescription,
        *,
        params: T_PARAMS = ...,
        headers: T_HEADERS = ...,
    ) -> PriceOverview:
        ...

    @overload
    async def fetch_price_overview(
        self,
        obj: str,
        app: App,
        *,
        params: T_PARAMS = ...,
        headers: T_HEADERS = ...,
    ) -> PriceOverview:
        ...

    @currency_required
    async def fetch_price_overview(
        self,
        obj: str | ItemDescription,
        app: App = None,
        *,
        params: T_PARAMS = {},
        headers: T_HEADERS = {},
    ) -> PriceOverview:
        """
        Fetch price data.

        .. note:: This request is rate limited by Steam.

        :param obj: `market hash name` ofr the `Steam` item or `ItemDescription`
        :param app:
        :param params: extra params to pass to url
        :param headers: extra headers to send with request
        :return: `PriceOverview` dict
        :raises EResultError:
        """

        if isinstance(obj, ItemDescription):
            name = obj.market_hash_name
            app = obj.app
        else:  # str
            name = obj

        params = {
            "country": self.country,
            "currency": self.currency,
            "market_hash_name": name,
            "appid": app.value,
            **params,
        }
        try:
            r = await self.session.get(STEAM_URL.MARKET / "priceoverview", params=params, headers=headers)
        except ClientResponseError as e:
            if e.status == 429:
                raise RateLimitExceeded("You have been rate limited, rest for a while!") from e
            else:
                raise e

        rj: PriceOverview = await r.json()
        success = EResult(rj.get("success"))
        if success is not EResult.OK:
            raise EResultError(rj.get("message", "Failed to fetch price overview"), success, rj)

        return rj

    @overload
    async def get_item_listings(
        self,
        obj: ItemDescription,
        *,
        query: str = ...,
        start: int = ...,
        count: int = ...,
        if_modified_since: datetime | str = ...,
        params: T_PARAMS = ...,
        headers: T_HEADERS = ...,
    ) -> T_MARKET_ITEM_LISTINGS_DATA:
        ...

    @overload
    async def get_item_listings(
        self,
        obj: str,
        app: App,
        *,
        query: str = ...,
        start: int = ...,
        count: int = ...,
        if_modified_since: datetime | str = ...,
        params: T_PARAMS = ...,
        headers: T_HEADERS = ...,
    ) -> T_MARKET_ITEM_LISTINGS_DATA:
        ...

    @currency_required
    async def get_item_listings(
        self,
        obj: str | ItemDescription,
        app: App = None,
        *,
        query: str = "",
        start: int = 0,
        count: int = LISTING_COUNT,
        if_modified_since: datetime | str = None,
        params: T_PARAMS = {},
        headers: T_HEADERS = {},
        _item_descriptions_map: T_SHARED_DESCRIPTIONS = None,
        _market_econ_items_map: dict[str, MarketListingItem] = None,
    ) -> T_MARKET_ITEM_LISTINGS_DATA:
        """
        Fetch item listings from market.

        .. note::
            * You can paginate by yourself passing `start` arg. or use `market_listings` method.
            * This request is rate limited by Steam. It is strongly recommended to use `if_modified_since`

        :param obj: `market hash name` ofr the `Steam` item or `ItemDescription`
        :param app:
        :param count: page size
        :param start: offset position
        :param query: raw search query
        :param if_modified_since: `If-Modified-Since` header value
        :param params: extra params to pass to url
        :param headers: extra headers to send with request
        :return: list of `MarketListing`, total listings count, datetime object when resource was last modified
        :raises EResultError: for ordinary reasons
        :raises RateLimitExceeded: when you hit rate limit
        :raises ResourceNotModified: when `if_modified_since` header passed and Steam response with 304 status code
        """

        if isinstance(obj, ItemDescription):
            name = obj.market_hash_name
            app = obj.app
        else:  # str
            name = obj

        base_url = STEAM_URL.MARKET / f"listings/{app.value}/{name}"
        params = {
            "filter": query,
            "query": "",  # as web browser
            "country": self.country,
            "currency": self.currency,
            "start": start,
            "count": count,
            "language": self.language,
            **params,
        }
        headers = {"Referer": str(base_url), **headers}
        if if_modified_since:
            if isinstance(if_modified_since, datetime):
                headers["If-Modified-Since"] = format_time(if_modified_since)
            else:  # str
                headers["If-Modified-Since"] = if_modified_since

        try:
            r = await self.session.get(base_url / "render/", params=params, headers=headers)
        except ClientResponseError as e:
            if e.status == 429:
                raise RateLimitExceeded("You have been rate limited, rest for a while!") from e
            else:
                raise e

        last_modified = parse_time(r.headers["Last-Modified"])

        if r.status == 304:  # not modified if header "If-Modified-Since" is provided
            raise ResourceNotModified(last_modified, parse_time(r.headers["Expires"]))

        rj: dict[str, int | dict[str, dict]] = await r.json()
        success = EResult(rj.get("success"))
        if success is not EResult.OK:
            raise EResultError(rj.get("message", "Failed to fetch item listings"), success, rj)

        if not rj["total_count"] or not rj["assets"]:
            return [], 0, last_modified

        if _item_descriptions_map is None:
            _item_descriptions_map = {}
        if _market_econ_items_map is None:
            _market_econ_items_map = {}

        self._parse_descriptions_from_market_assets(rj["assets"], _item_descriptions_map)
        # Do we need to share items?
        self._parse_market_listing_items(rj["assets"], _item_descriptions_map, _market_econ_items_map)

        return self._create_market_listings(rj, _market_econ_items_map), rj["total_count"], last_modified

    @classmethod
    def _create_market_listings(cls, data: dict, items_map: dict[str, MarketListingItem]) -> list[MarketListing]:
        return [
            MarketListing(
                id=int(l_data["listingid"]),
                item=items_map[
                    create_ident_code(
                        l_data["asset"]["id"],
                        l_data["asset"]["contextid"],
                        l_data["asset"]["appid"],
                    )
                ],
                currency=Currency(int(l_data["currencyid"]) - 2000),
                price=int(l_data["price"]),
                fee=int(l_data["fee"]),
                converted_currency=Currency(int(l_data["converted_currencyid"]) - 2000)
                if "converted_currencyid" in l_data
                else None,
                converted_fee=int(l_data["converted_fee"]) if "converted_fee" in l_data else None,
                converted_price=int(l_data["converted_price"]) if "converted_price" in l_data else None,
            )
            for l_data in data["listinginfo"].values()
            # due to "0", ignore items with no amount and prices (supposedly purchased)
            if int(l_data["asset"]["amount"])
        ]

    @classmethod
    def _parse_descriptions_from_market_assets(
        cls,
        assets: dict[str, dict[str, dict[str, dict]]],
        item_descriptions_map: dict[str, ItemDescription],
    ):
        for app_id, app_data in assets.items():
            for context_id, context_data in app_data.items():
                for asset_id, mixed_data in context_data.items():  # asset+descr data
                    key = create_ident_code(mixed_data["instanceid"], mixed_data["classid"], app_id)
                    item_descriptions_map[key] = cls._create_item_descr(mixed_data)

    @staticmethod
    def _parse_market_listing_items(
        data: dict[str, dict[str, dict[str, dict]]],
        item_descriptions_map: dict[str, ItemDescription],
        items_map: dict[str, MarketListingItem],
    ):
        for app_id, app_data in data.items():
            for context_id, context_data in app_data.items():
                for a_data in context_data.values():
                    key = create_ident_code(a_data["id"], context_id, app_id)
                    if key not in items_map:
                        items_map[key] = MarketListingItem(
                            asset_id=int(a_data["id"]),
                            market_id=0,  # set in market listing post init
                            unowned_id=int(a_data["unowned_id"]),
                            unowned_context_id=int(a_data["unowned_contextid"]),
                            app_context=AppContext((App(int(a_data["appid"])), int(a_data["contextid"]))),
                            description=item_descriptions_map[
                                create_ident_code(
                                    a_data["instanceid"],
                                    a_data["classid"],
                                    app_id,
                                )
                            ],
                        )

    # without async for proper type hinting in VsCode and PyCharm at least with `async for`
    @overload
    def market_listings(
        self,
        obj: ItemDescription,
        *,
        query: str = ...,
        start: int = ...,
        count: int = ...,
        if_modified_since: datetime | str = ...,
        params: T_PARAMS = ...,
        headers: T_HEADERS = ...,
    ) -> AsyncIterator[T_MARKET_ITEM_LISTINGS_DATA]:
        ...

    @overload
    def market_listings(
        self,
        obj: str,
        app: App,
        *,
        query: str = ...,
        start: int = ...,
        count: int = ...,
        if_modified_since: datetime | str = ...,
        params: T_PARAMS = ...,
        headers: T_HEADERS = ...,
    ) -> AsyncIterator[T_MARKET_ITEM_LISTINGS_DATA]:
        ...

    async def market_listings(
        self,
        obj: str | ItemDescription,
        app: App = None,
        *,
        query: str = "",
        start: int = 0,
        count: int = LISTING_COUNT,
        if_modified_since: datetime | str = None,
        params: T_PARAMS = {},
        headers: T_HEADERS = {},
        _item_descriptions_map: T_SHARED_DESCRIPTIONS = None,
        _market_econ_items_map: dict[str, MarketListingItem] = None,
    ) -> AsyncIterator[T_MARKET_ITEM_LISTINGS_DATA]:
        """
        Fetch item listings from market. Return async iterator to paginate over listings pages.

        .. note:: This request is rate limited by Steam. It is strongly recommended to use `if_modified_since`

        :param obj: `market hash name` ofr the `Steam` item or `ItemDescription`
        :param app:
        :param count: page size
        :param start: offset position
        :param query: raw search query
        :param if_modified_since: `If-Modified-Since` header value
        :param params: extra params to pass to url
        :param headers: extra headers to send with request
        :return: `AsyncIterator` that yields list of `MarketListing`, total listings count, datetime object when resource was last modified
        :raises EResultError: for ordinary reasons
        :raises RateLimitExceeded: when you hit rate limit
        """

        if _item_descriptions_map is None:
            _item_descriptions_map = {}
        if _market_econ_items_map is None:
            _market_econ_items_map = {}

        total_count: int = 10e6  # simplify logic for initial iteration
        while total_count > start:
            # browser loads first batch from document request and not json api point, but anyway
            # avoid excess destructuring
            listings_data = await self.get_item_listings(
                obj,
                app,
                query=query,
                count=count,
                if_modified_since=if_modified_since,
                params=params,
                headers=headers,
                _item_descriptions_map=_item_descriptions_map,
                _market_econ_items_map=_market_econ_items_map,
                start=start,
            )
            total_count = listings_data[1]
            start += count

            yield listings_data

    @overload
    async def get_item_name_id(self, obj: ItemDescription, *, headers: T_HEADERS = ...) -> int:
        ...

    @overload
    async def get_item_name_id(self, obj: str, app: App, *, headers: T_HEADERS = ...) -> int:
        ...

    async def get_item_name_id(self, obj: str | ItemDescription, app: App = None, *, headers: T_HEADERS = {}) -> int:
        """
        Fetch item from `Steam Community Market` page, find and return `item_nameid`

        :param obj: `market hash name` ofr the `Steam` item or `ItemDescription`
        :param app:
        :param headers: extra headers to send with request
        :return: `item_nameid`
        :raises RateLimitExceeded: when you hit rate limit

        .. seealso:: https://github.com/somespecialone/steam-item-name-ids
        """

        if isinstance(obj, ItemDescription):
            url = obj.market_url
        else:  # str
            url = STEAM_URL.MARKET / f"listings/{app.value}/{obj}"

        try:
            res = await self.session.get(url, headers=headers)
        except ClientResponseError as e:
            if e.status == 429:
                raise RateLimitExceeded("You have been rate limited, rest for a while!") from e
            else:
                raise e

        text = await res.text()

        return find_item_nameid_in_text(text)

    async def get_market_search_app_filters(self, app: App) -> dict:
        """
        Fetch app filters facets for market search.
        You can see them when click on `Show advanced options...`
        button under search input field on `Steam` market page
        """

        r = await self.session.get(
            STEAM_URL.MARKET / f"appfilters/{app.value}",
            headers={"Referer": str(STEAM_URL.MARKET)},
        )
        rj = await r.json()
        success = EResult(rj.get("success"))
        if success is not EResult.OK:
            raise EResultError(rj.get("message", "Failed to get app filters for market search"), success, rj)

        return rj

    async def get_market_search_results(
        self,
        query="",
        app: App = None,
        *,
        start=0,
        count=10,
        descriptions=False,
        sort_column: T_SORT_COLUMN = "default",
        sort_dir: T_SORT_DIR = "desc",
        filters: str | Mapping[str, str | Sequence[str]] = "",
        headers: T_HEADERS = {},
        _item_descriptions_map: T_SHARED_DESCRIPTIONS = None,
    ) -> tuple[list[MarketSearchItem], int]:
        """
        Request search results from `Steam` for market.

        You can find how to write filters by investigating how browser sends requests to a
        `https://steamcommunity.com/market/search/render/` endpoint with enabled at least one option from
        `advanced options` window.

        Example for `CS2`, with `Collection` as `The Anubis Collection` will look like:
            * get_market_search_results(app=App.CS2, filters={"category_730_ItemSet[]": "tag_set_anubis"})

        :param query: raw search query
        :param app: just `Steam` app
        :param start: start result position
        :param count: total count of results on page
        :param descriptions: to search in descriptions
        :param sort_column: column to sort by
        :param sort_dir: direction to sort by
        :param filters: app search filters for market
        :param headers: extra headers to send with request
        :return: list of `MarketSearchItem`, total results count
        :raises EResultError: for ordinary reasons
        :raises RateLimitExceeded: when you hit rate limit
        """

        req_params = {
            "norender": 1,
            "query": query,  # empty, like in browser's request
            "start": start,
            "count": count,
            "sort_column": sort_column,
            "sort_dir": sort_dir,
            "search_descriptions": to_int_boolean(descriptions),
        }

        referer_params = {
            "q": query,
            "sort_column": sort_column,
            "sort_dir": sort_dir,
        }

        if app is not None:
            req_params["appid"] = referer_params["appid"] = app.value
        if descriptions:
            referer_params["descriptions"] = 1

        referer = SEARCH_URL % referer_params % filters

        try:
            r = await self.session.get(
                SEARCH_RENDER_URL % req_params % filters,
                headers={"Referer": str(referer), **headers},
            )
        except ClientResponseError as e:
            if e.status == 429:
                raise RateLimitExceeded("You have been rate limited, rest for a while!") from e
            else:
                raise e

        rj: dict[str, int | list[dict[str, str | int | dict[str, str | int]]]] = await r.json()
        success = EResult(rj.get("success"))
        if success is not EResult.OK:
            raise EResultError(rj.get("message", "Failed to fetch market search results"), success, rj)

        if not rj["total_count"] or not rj["results"]:
            return [], 0

        if _item_descriptions_map is None:
            _item_descriptions_map = {}

        items = []
        for item_data in rj["results"]:
            descr_ident_code = create_ident_code(
                item_data["asset_description"]["instanceid"],
                item_data["asset_description"]["classid"],
                item_data["asset_description"]["appid"],
            )

            if descr_ident_code in _item_descriptions_map:
                description = _item_descriptions_map[descr_ident_code]
            else:
                description = self._create_item_descr(item_data["asset_description"])
                _item_descriptions_map[descr_ident_code] = description

            item = MarketSearchItem(
                sell_listings=item_data["sell_listings"],
                sell_price=item_data["sell_price"],
                sell_price_text=item_data["sell_price_text"],
                sale_price_text=item_data["sale_price_text"],
                app_icon=item_data["app_icon"],
                app_name=item_data["app_name"],
                description=description,
            )
            items.append(item)

        return items, rj["total_count"]

    async def market_search_results(
        self,
        query="",
        app: App = None,
        *,
        start=0,
        count=10,
        descriptions=False,
        sort_column: T_SORT_COLUMN = "default",
        sort_dir: T_SORT_DIR = "desc",
        filters: str | Mapping[str, str | Sequence[str]] = "",
        headers: T_HEADERS = {},
        _item_descriptions_map: T_SHARED_DESCRIPTIONS = None,
    ) -> AsyncIterator[tuple[list[MarketSearchItem], int]]:
        """
        Request search results from `Steam` for market.
        Return async iterator to paginate over market search results pages.

        You can find how to write filters by investigating how browser sends requests to a
        `https://steamcommunity.com/market/search/render/` endpoint with enabled at least one option from
        `advanced options` window.

        Example for `CS2`, with `Collection` as `The Anubis Collection` will look like:
            * get_market_search_results(..., **{"category_730_ItemSet[]": "tag_set_anubis"})

        :param query: raw search query
        :param app: just `Steam` app
        :param start: start result position
        :param count: total count of results on page
        :param descriptions: to search in descriptions
        :param sort_column: column to sort by
        :param sort_dir: direction to sort by
        :param filters: app search filters for market
        :param headers: extra headers to send with request
        :return: `AsyncIterator` that yields list of `MarketSearchItem`, total results count
        :raises EResultError: for ordinary reasons
        :raises RateLimitExceeded: when you hit rate limit
        """

        if _item_descriptions_map is None:
            _item_descriptions_map = {}

        total_count: int = 10e6  # simplify logic for initial iteration
        while total_count > start:
            # browser loads first batch from document request and not json api point, but anyway
            # avoid excess destructuring
            search_results = await self.get_market_search_results(
                query,
                app,
                count=count,
                descriptions=descriptions,
                sort_column=sort_column,
                sort_dir=sort_dir,
                headers=headers,
                _item_descriptions_map=_item_descriptions_map,
                start=start,
                filters=filters,
            )
            total_count = search_results[1]
            start += count

            yield search_results



"""
Trade and interact with steam market, webapi, guard.
"""

from .exceptions import *
from .constants import (
    App,
    AppContext,
    STEAM_URL,
    Currency,
    Language,
    TradeOfferStatus,
    MarketListingStatus,
    EResult,
    ConfirmationType,
)
from .client import SteamClient, SteamPublicClient
from .models import (
    MarketListing,
    ItemDescription,
    EconItem,
    Confirmation,
    MarketListingItem,
    MyMarketListing,
    BuyOrder,
    MarketHistoryListing,
    MarketHistoryListingItem,
    MarketHistoryEvent,
    TradeOfferItem,
    TradeOffer,
    HistoryTradeOffer,
    HistoryTradeOfferItem,
    ItemOrdersHistogram,
)



from re import search as re_search
from json import loads
from typing import AsyncIterator, overload, Callable, final, TYPE_CHECKING

from aiohttp import ClientSession
from aiohttp.client import _RequestContextManager

from .constants import STEAM_URL, Currency, AppContext, Language, T_PARAMS, T_HEADERS, EResult
from .typed import WalletInfo, FundWalletInfo
from .exceptions import EResultError, SessionExpired, SteamError
from .utils import account_id_to_steam_id, generate_device_id
from .models import Notifications, EconItem

from .mixins.public import SteamCommunityPublicMixin, INV_COUNT, INV_ITEM_DATA, T_SHARED_DESCRIPTIONS
from .mixins.profile import ProfileMixin
from .mixins.trade import TradeMixin
from .mixins.market import MarketMixin


DEF_COUNTRY = "UA"
DEF_TZ_OFFSET = "10800,0"


class SteamPublicClientBase(SteamCommunityPublicMixin):
    __slots__ = ()

    @overload
    def __init__(
        self,
        *,
        language: Language = ...,
        tz_offset: str = ...,
        currency: Currency = ...,
        country: str = ...,
        session: ClientSession = ...,
        user_agent: str = ...,
    ):
        ...

    @overload
    def __init__(
        self,
        *,
        language: Language = ...,
        tz_offset: str = ...,
        currency: Currency = ...,
        country: str = ...,
        proxy: str = ...,
        user_agent: str = ...,
    ):
        ...

    def __init__(
        self,
        *,
        language=Language.ENGLISH,
        tz_offset=DEF_TZ_OFFSET,
        currency=Currency.USD,
        country=DEF_COUNTRY,
        session: ClientSession = None,
        proxy: str = None,
        user_agent: str = None,
    ):
        """
        Base `Steam Public Community` client class.
        Implements construction method.

        .. note:: Subclass this if you want to make your custom public client

        :param language: language of `Steam` descriptions, responses, etc... Will be set to a cookie
        :param currency: currency of market data in `Steam` responses
        :param country: country of market data in `Steam` responses
        :param tz_offset: timezone offset. Will be set to a cookie
        :param session: session instance. Must be created with `raise_for_status=True` for client to work properly
        :param proxy: proxy url as string. Can be in format `scheme://username:password@host:port` or `scheme://host:port`
        :param user_agent: user agent header value. Strongly advisable to set this
        """

        self.session = self._session_helper(session, proxy)

        if user_agent:
            self.user_agent = user_agent

        self.language = language
        self.tz_offset = tz_offset
        self.currency = currency
        self.country = country

    def __repr__(self) -> str:
        return f"{type(self).__name__}(language={self.language}, tz_offset={self.tz_offset}, currency={self.currency}, country={self.country})"


class SteamClientBase(SteamPublicClientBase, ProfileMixin, MarketMixin, TradeMixin):
    __slots__ = ()

    @overload
    def __init__(
        self,
        steam_id: int,
        username: str,
        password: str,
        shared_secret: str,
        identity_secret: str = ...,
        *,
        access_token: str = ...,
        refresh_token: str = ...,
        api_key: str = ...,
        trade_token: str = ...,
        language: Language = ...,
        wallet_currency: Currency = ...,
        wallet_country: str = ...,
        tz_offset: str = ...,
        session: ClientSession = ...,
        user_agent: str = ...,
    ):
        ...

    @overload
    def __init__(
        self,
        steam_id: int,
        username: str,
        password: str,
        shared_secret: str,
        identity_secret: str = ...,
        *,
        access_token: str = ...,
        refresh_token: str = ...,
        api_key: str = ...,
        trade_token: str = ...,
        language: Language = ...,
        wallet_currency: Currency = ...,
        wallet_country: str = ...,
        tz_offset: str = ...,
        proxy: str = ...,
        user_agent: str = ...,
    ):
        ...

    def __init__(
        self,
        # It is possible to get steam id from the cookie and then arg will not be necessary,
        # but typical use case of the library means that the user already knows the steam id
        steam_id: int,  # first as intend to be mandatory
        username: str,
        password: str,
        shared_secret: str,
        identity_secret: str = None,
        *,
        access_token: str = None,
        refresh_token: str = None,
        api_key: str = None,
        trade_token: str = None,
        language=Language.ENGLISH,
        wallet_currency: Currency = None,
        wallet_country=DEF_COUNTRY,
        tz_offset=DEF_TZ_OFFSET,
        session: ClientSession = None,
        proxy: str = None,
        user_agent: str = None,
    ):
        """
        Base `Steam Community` client class.
        Implements construction, util, account (inventory, wallet) methods.

        .. note:: Subclass this if you want to make your custom client

        :param steam_id: steam id (id64) or account id (id32)
        :param username:
        :param password:
        :param shared_secret:
        :param identity_secret: required for confirmations
        :param access_token: encoded JWT token string
        :param refresh_token: encoded JWT token string
        :param api_key: `Steam Web API` key to have access to `Steam Web API`. Can be used instead of `access_token`
        :param trade_token: trade token of account. Needed to create a trade url
        :param language: language of `Steam` descriptions, responses, etc... Will be set to a cookie
        :param wallet_currency: currency of account wallet. Market methods requires this to be set
        :param wallet_country: country of account wallet. Market methods requires this to be set
        :param tz_offset: timezone offset. Will be set to a cookie
        :param session: session instance. Must be created with `raise_for_status=True` for client to work properly
        :param proxy: proxy url as string. Can be in format `scheme://username:password@host:port` or `scheme://host:port`
        :param user_agent: user agent header value. Strongly advisable to set this
        """

        super().__init__(
            language=language,
            tz_offset=tz_offset,
            currency=wallet_currency,
            country=wallet_country,
            session=session,
            proxy=proxy,
            user_agent=user_agent,
        )

        # guard
        self.steam_id = account_id_to_steam_id(steam_id) if steam_id < 4294967296 else steam_id  # steam id64
        self.device_id = generate_device_id(self.steam_id)

        self._shared_secret = shared_secret
        self._identity_secret = identity_secret

        # login
        self.username = username
        self._password = password
        self.access_token = access_token
        self.refresh_token = refresh_token

        # profile
        self.trade_token = trade_token

        # web api
        self._api_key = api_key

    # aliases for convenience
    @property
    def wallet_currency(self) -> Currency:
        return self.currency

    @property
    def wallet_country(self) -> str:
        return self.country

    async def prepare(self, api_key_domain: str = None, *, force=False):
        """
        Prepares client to work by loading main attributes (trade token, currency and country, optionally api key)
        from `Steam`. Register trade token and api key (optionally) if there is none.
        Edit privacy settings of inventory and related staff to be public

        :param api_key_domain: domain to register `Steam Web Api` key.
            If not passed, api key will not be fetched and registered.
        :param force: force to reload all data even if it presented on client
        """

        if (not self._api_key or force) and api_key_domain:
            await self.get_api_key()
            not self._api_key and await self.register_new_api_key(api_key_domain)

        if not self.trade_token or force:
            await self.get_trade_token()
            not self.trade_token and await self.register_new_trade_url()

        if (not self.currency or not self.country) or force:
            wallet_info = await self.get_wallet_info()
            self.country = wallet_info["wallet_country"]
            self.currency = Currency(wallet_info["wallet_currency"])

        # avoid unnecessary privacy editing
        profile_data = await self.get_profile_data()
        if (
            profile_data["Privacy"]["PrivacySettings"]["PrivacyInventory"] != 3
            or profile_data["Privacy"]["PrivacySettings"]["PrivacyInventoryGifts"] != 3
            or profile_data["Privacy"]["PrivacySettings"]["PrivacyProfile"] != 3
        ):
            await self.edit_privacy_settings(inventory=3, inventory_gifts=True, profile=3)

    async def get_wallet_info(self) -> WalletInfo:
        """
        Fetch wallet info from inventory page.

        .. note:: May reset new items notifications count.

        :return: wallet info
        :raises EResultError: for ordinary reasons
        """

        r = await self.session.get(self.profile_url / "inventory", headers={"Referer": str(self.profile_url)})
        rt = await r.text()
        info: WalletInfo = loads(re_search(r"g_rgWalletInfo = (?P<info>.+);", rt)["info"])
        success = EResult(info.get("success"))
        if success is not EResult.OK:
            raise EResultError(info.get("message", "Failed to fetch wallet info from inventory"), success, info)

        return info

    async def get_wallet_balance(self) -> int:
        """
        Fetch wallet info from inventory page, parse and return balance.

        .. note:: May reset new items notifications count.

        :return: wallet balance as integer
        :raises EResultError: for ordinary reasons
        """

        info = await self.get_wallet_info()
        return int(info["wallet_balance"])

    async def get_fund_wallet_info(self) -> FundWalletInfo:
        """
        Fetch fund wallet info from `Steam Store` domain.

        .. note:: `Steam Store` domain uses own and different one access token from `Steam Community`,
            so it can be expired easily as it not used much in `aiosteampy`

        :return: `FundWalletInfo`
        :raises EResultError: for ordinary reasons
        """

        r = await self.session.get(STEAM_URL.STORE / "api/getfundwalletinfo")
        rj: FundWalletInfo = await r.json()
        success = EResult(rj.get("success"))
        if success is not EResult.OK:
            raise EResultError(rj.get("message", "Failed to fetch wallet balance"), success, rj)

        return rj

    async def get_notifications(self) -> Notifications:
        """Get notifications count."""

        headers = {"Referer": str(self.profile_url)}
        r = await self.session.get(STEAM_URL.COMMUNITY / "actions/GetNotificationCounts", headers=headers)
        rj = await r.json()
        return Notifications(*(rj["notifications"][str(i)] for i in range(1, 12) if i != 7))

    # https://github.com/DoctorMcKay/node-steamcommunity/blob/7c564c1453a5ac413d9312b8cf8fe86e7578b309/index.js#L275
    def reset_items_notifications(self) -> _RequestContextManager:
        """Fetches self inventory page, which resets new items notifications to 0."""

        return self.session.get(self.profile_url / "inventory/")

    async def get_inventory(
        self,
        app_context: AppContext,
        *,
        last_assetid: int = None,
        count=INV_COUNT,
        params: T_PARAMS = {},
        headers: T_HEADERS = {},
        _item_descriptions_map: T_SHARED_DESCRIPTIONS = None,
    ) -> INV_ITEM_DATA:
        """
        Fetches self inventory.

        .. note::
            * You can paginate by yourself passing `last_assetid` arg
            * `count` arg value that less than 2000 lead to responses with strange amount of assets

        :param app_context: `Steam` app+context
        :param last_assetid:
        :param count: page size
        :param params: extra params to pass to url
        :param headers: extra headers to send with request
        :return: list of `EconItem`, total count of items in inventory, last asset id of the list
        :raises EResultError: for ordinary reasons
        :raises RateLimitExceeded: when you hit rate limit
        :raises SessionExpired:
        """

        try:
            return await self.get_user_inventory(
                self.steam_id,
                app_context,
                last_assetid=last_assetid,
                count=count,
                params=params,
                headers=headers,
                _item_descriptions_map=_item_descriptions_map,
            )
        except SteamError as e:
            if "private" in e.args[0]:  # self inventory can't be private
                raise SessionExpired from e
            else:
                raise e

    def inventory(
        self,
        app_context: AppContext,
        *,
        last_assetid: int = None,
        count=INV_COUNT,
        params: T_PARAMS = {},
        headers: T_HEADERS = {},
        _item_descriptions_map: T_SHARED_DESCRIPTIONS = None,
    ) -> AsyncIterator[INV_ITEM_DATA]:
        """
        Fetches self inventory. Return async iterator to paginate over inventory pages.

        .. note:: `count` arg value that less than 2000 lead to responses with strange amount of assets

        :param app_context: `Steam` app+context
        :param last_assetid:
        :param count: page size
        :param params: extra params to pass to url
        :param headers: extra headers to send with request
        :return: `AsyncIterator` that yields list of `EconItem`, total count of items in inventory, last asset id of the list
        :raises EResultError: for ordinary reasons
        :raises RateLimitExceeded: when you hit rate limit
        :raises SessionExpired:
        """

        return self.user_inventory(
            self.steam_id,
            app_context,
            last_assetid=last_assetid,
            count=count,
            params=params,
            headers=headers,
            _item_descriptions_map=_item_descriptions_map,
        )

    @overload
    async def get_inventory_item(
        self,
        app_context: AppContext,
        obj: int = ...,
        *,
        params: T_PARAMS = ...,
        headers: T_HEADERS = ...,
        **item_attrs,
    ) -> EconItem | None:
        ...

    @overload
    async def get_inventory_item(
        self,
        app_context: AppContext,
        obj: Callable[[EconItem], bool],
        *,
        params: T_PARAMS = ...,
        headers: T_HEADERS = ...,
    ) -> EconItem | None:
        ...

    async def get_inventory_item(
        self,
        app_context: AppContext,
        obj: int | Callable[[EconItem], bool] = None,
        *,
        params: T_PARAMS = {},
        headers: T_HEADERS = {},
        _item_descriptions_map: T_SHARED_DESCRIPTIONS = None,
        **item_attrs,
    ) -> EconItem | None:
        """
        Fetch and iterate over inventory item pages of self until find one that satisfies passed arguments.

        :param app_context: `Steam` app+context
        :param obj: asset id or predicate function
        :param params: extra params to pass to url
        :param headers: extra headers to send with request
        :param item_attrs: additional item attributes and values
        :return: `EconItem` or `None`
        :raises EResultError: for ordinary reasons
        :raises RateLimitExceeded: when you hit rate limit
        :raises SessionExpired:
        """

        try:
            return await self.get_user_inventory_item(
                self.steam_id,
                app_context,
                obj,
                params=params,
                headers=headers,
                _item_descriptions_map=_item_descriptions_map,
                **item_attrs,
            )
        except SteamError as e:
            if "private" in e.args[0]:  # self inventory can't be private
                raise SessionExpired from e
            else:
                raise e

    def __repr__(self) -> str:
        return f"{type(self).__name__}(steam_id={self.steam_id}, username={self.username})"


@final
class SteamClient(SteamClientBase):
    __slots__ = (
        "_refresh_token",
        "session",
        "username",
        "steam_id",
        "_password",
        "_shared_secret",
        "_identity_secret",
        "_api_key",
        "trade_token",
        "device_id",
        "currency",
        "country",
    )

    if TYPE_CHECKING:  # for PyCharm pop-up

        @overload
        def __init__(
            self,
            steam_id: int,
            username: str,
            password: str,
            shared_secret: str,
            identity_secret: str = ...,
            *,
            access_token: str = ...,
            refresh_token: str = ...,
            api_key: str = ...,
            trade_token: str = ...,
            language: Language = ...,
            wallet_currency: Currency = ...,
            wallet_country: str = ...,
            tz_offset: str = ...,
            session: ClientSession = ...,
            user_agent: str = ...,
        ):
            ...

        @overload
        def __init__(
            self,
            steam_id: int,
            username: str,
            password: str,
            shared_secret: str,
            identity_secret: str = ...,
            *,
            access_token: str = ...,
            refresh_token: str = ...,
            api_key: str = ...,
            trade_token: str = ...,
            language: Language = ...,
            wallet_currency: Currency = ...,
            wallet_country: str = ...,
            tz_offset: str = ...,
            proxy: str = ...,
            user_agent: str = ...,
        ):
            ...

        def __init__(
            self,
            steam_id: int,
            username: str,
            password: str,
            shared_secret: str,
            identity_secret: str = None,
            *,
            access_token: str = None,
            refresh_token: str = None,
            api_key: str = None,
            trade_token: str = None,
            language=Language.ENGLISH,
            wallet_currency: Currency = None,
            wallet_country=DEF_COUNTRY,
            tz_offset=DEF_TZ_OFFSET,
            session: ClientSession = None,
            proxy: str = None,
            user_agent: str = None,
        ):
            """
            Ready to use client class with inherited methods from all mixins. Must be logged in.

            :param steam_id: steam id (id64) or account id (id32)
            :param username:
            :param password:
            :param shared_secret:
            :param identity_secret: required for confirmations
            :param access_token: encoded JWT token string
            :param refresh_token: encoded JWT token string
            :param api_key: `Steam Web API` key to have access to `Steam Web API`. Can be used instead of `access_token`
            :param trade_token: trade token of account. Needed to create a trade url
            :param language: language of `Steam` descriptions, responses, etc... Will be set to a cookie
            :param wallet_currency: currency of account wallet. Market methods requires this to be set
            :param wallet_country: country of account wallet. Market methods requires this to be set
            :param tz_offset: timezone offset. Will be set to a cookie
            :param session: session instance. Must be created with `raise_for_status=True` for client to work properly
            :param proxy: proxy url as string. Can be in format `scheme://username:password@host:port` or `scheme://host:port`
            :param user_agent: user agent header value. Strongly advisable to set this
            """


@final
class SteamPublicClient(SteamPublicClientBase):
    __slots__ = ("session", "currency", "country")

    if TYPE_CHECKING:  # for PyCharm pop-up

        @overload
        def __init__(
            self,
            *,
            language: Language = ...,
            tz_offset: str = ...,
            currency: Currency = ...,
            country: str = ...,
            session: ClientSession = ...,
            user_agent: str = ...,
        ):
            ...

        @overload
        def __init__(
            self,
            *,
            language: Language = ...,
            tz_offset: str = ...,
            currency: Currency = ...,
            country: str = ...,
            proxy: str = ...,
            user_agent: str = ...,
        ):
            ...

        def __init__(
            self,
            *,
            language=Language.ENGLISH,
            tz_offset=DEF_TZ_OFFSET,
            currency=Currency.USD,
            country=DEF_COUNTRY,
            session: ClientSession = None,
            proxy: str = None,
            user_agent: str = None,
        ):
            """
            Client contains public methods that do not require authentication.

            :param language: language of `Steam` descriptions, responses, etc... Will be set to a cookie
            :param currency: currency of market data in `Steam` responses
            :param country: country of market data in `Steam` responses
            :param tz_offset: timezone offset. Will be set to a cookie
            :param session: session instance. Must be created with `raise_for_status=True` for client to work properly
            :param proxy: proxy url as string. Can be in format `scheme://username:password@host:port` or `scheme://host:port`
            :param user_agent: user agent header value. Strongly advisable to set this
            """
